{********************************************************************}
{                         C D D E V . P A S                          }
{*------------------------------------------------------------------*}
{ Aufgabe        : Utils fr CD-Device-Treiber Programmierung        }
{*------------------------------------------------------------------*}
{ Autor          : Michael Tischer / Bruno Jennrich                  }
{ entwickelt am  : 08.04.1994                                        }
{ letztes Update :  8.10.1994                                        }
{********************************************************************}

{$A-}                           { kein Word-Alignment von Strukturen }
{$X+}   { erweiterte Syntax, Auswertung von Fkt-Ergebnissen optional }

Unit CDDEV;

Interface
Uses CDUTIL, CRT;

Const
 DEV_ERROR = ( 1 shl 15 );        { Gesetzt, dann in Bits 0-7 Fehler }
 DEV_BUSY  = ( 1 shl 9 );                   { Device ist beschftigt }
 DEV_DONE  = ( 1 shl 8 );      { Device hat letzte Operation beendet }

{- Bei gesetztem Fehlerbit folgende Fehlermeldungen in Low-Byte -----}
 DEV_WRITE_PROTECT_VIOLATION   = 0;
 DEV_UNKNOWN_UNIT              = 1;
 DEV_DRIVE_NOT_READY           = 2;
 DEV_UNKNOWN_COMMAND           = 3;
 DEV_CRC_ERROR                 = 4;
 DEV_BAD_DRIVE_REQUEST_LEN     = 5;
 DEV_SEEK_ERROR                = 6;
 DEV_UNKNOWN_MEDIA             = 7;
 DEV_SECTOR_NOT_FOUND          = 8;
 DEV_PRINTER_OUT_OF_PAPER      = 9;
 DEV_WRITE_FAULT               = 10;
 DEV_READ_FAULT                = 11;
 DEV_GENERAL_FAILURE           = 12;
 DEV_RESERVED1                 = 13;
 DEV_RESERVED2                 = 14;
 DEV_INVALID_DISK_CHANGE       = 15;

{- Device-Kommandos -------------------------------------------------}
 DEVCMD_INIT = 0;    { INIT-Kommando wird nur einmal whrend des  }
                     { Bootens aufgerufen, danach nicht mehr.     }
 DEVCMD_IOCTL_READ = 3;

{-- Liefert Adresse des DeviceHeaders im RAM ---}
Const IOCTLI_GET_DEVHDR_ADRESS = 0;
Type MSCDEX_Raddr = Record 
       bControlBlockCode    : Byte;                            { = 0 }
       lpDeviceHeaderAdress : ^DeviceHeader;  { DeviceHeader-Adresse }
     End;

{-- Ermittelt Schreib-/Lesekopf-Position -------------------------}
Const IOCTLI_GET_HEAD_LOCATION = 1;
Type MSCDEX_LocHead = Record
       bControlBlockCode,                                      { = 1 }
       bAdressingMode    : Byte;        { = HSG (0) oder REDBOOK (1) }
       lLocation         : Longint;        { Position des (Schreib-) }
    End;                                       { Lesekopfes (Sektor) }

{-- Reserviertes Kommando ------------------------------------------}
Const IOCTLI_RESERVED1 = 2;

{-- Ermittelt Fehlerstatistik (noch nicht spezifiziert). -----------}
Const IOCTLI_ERROR_STATISTICS = 3;
Type MSCDEX_ErrStat = Record
       bControlBlockCode : Byte;                               { = 3 }
       bErrStat          : Array[0..0] of Byte;   { Gre des Arrays }
     End;                                     { noch nicht definiert }

{-- Ermittelt, welche CD-Kanle auf welchem CD-Player Ausgang -------}
{-- liegen und wie laut diese sind.                           -------}
Const IOCTLI_GET_AUDIO_CHANNEL_INFO = 4;
Type MSCDEX_AudInfo = Record
       bControlBlockCode,                                      { = 4 }
       bOutput0, bVolume0,     { Welche Audio-Spuren werden an       }
       bOutput1, bVolume1,     { welchen CD-Ausgang umgeleitet, und  }
       bOutput2, bVolume2,     { wie laut sind die CD-Ausgnge?      }
       bOutput3, bVolume3 : Byte
    End;

{-- Lese Daten direkt vom Laufwerk (Daten sind Hardwareabhngig) ---}
Const IOCTLI_READ_DRIVE_BYTES = 5;
Type MSCDEX_DrvBytes = Record
       bControlBlockCode,                                      { = 5 }
       bNumReadBytes     : Byte;        { = Anzahl zu lesender Bytes }
       bReadBuffer       : Array[0..127] of Byte;
     End;         { In 128-Byte Blcken (vorherige werden verworfen) }

{-- Ermittle Laufwerkszustand ---------------------------------------}
Const IOCTLI_GET_DEV_STAT = 6;
Type MSCDEX_DevStat = Record
       bControlBlockCode : Byte;                               { = 6 }
       lDeviceStatus     : Longint;        { enthlt nach Aufruf den }
    End;                                          { Laufwerks-Status }

Const
   {- Lade geffnet -------------------------------------------------}
   DS_DOOR_OPEN                = $00000001;

   {- Lade nicht veriegelt ------------------------------------------}
   DS_DOOR_UNLOCKED            = $00000002;

   {- Daten knnen Cooked und Raw gelesen werden, sonst nur Cooked --}
   DS_READS_COOKED_AND_RAW     = $00000004;

   {- Lauwerk kann auch schreiben (z.B. WORM), sonst nur lesen ------}
   DS_READ_WRITE               = $00000008;

   {- Laufwerk untersttzt Musik- und Video CD's, sonst nur Daten ---}
   DS_AUDIO_VIDEO              = $00000010;

   DS_ISO_9600_INTERLEAVE      = $00000020;       { Bit 6 reserviert }

   {- vorrausschauendes Lesen mglich -------------------------------}
   DS_PREFETCH                 = $00000080;

   {- Audio-Kanle knnen an CD-Ausgnge umgeleitet werden ----------}
   DS_MANIPULATE_AUDIO_CHANNEL = $00000100;

   {- Adressierungsmodi HSG und REDBOOK, sonst nur HSG --------------}
   DS_HSG_AND_REDBOOK          = $00000200;

   {- Keine Disk im Laufwerk ----------------------------------------}
   DS_NO_DISC_IN_DRIVE         = $00000800;

   {- Sub-Channel-Info kann gelesen werden --------------------------}
   DS_SUPPORT_RW_SUB_CHANNELS  = $00001000;

{-- Ermittelt Sektor-Gre in Bytes ---------------------------------}
Const IOCTLI_SECTOR_SIZE = 7;
Type MSCDEX_SectSize = Record
       bControlBlockCode,                                      { = 7 }
       bReadMode         : Byte;                   { COOKED oder RAW }
       wSectorSize       : Word;  { enthlt nach AUfruf Sektor-Gre }
     End;                         { COOKED: 2048 oder RAW: 2352 Byte }

{-- Position des Lead-Out-Tracks (erster nach Ende der CD) ----------}
Const IOCTLI_VOLUME_SIZE = 8;
Type MSCDEX_VolSize = Record
       bControlBlockCode : Byte;                               { = 8 }
       lVolumeSize       : Longint;{ nach Aufruf Position des letzten}
     End;                                                {Tracks + 1 }

{-- Hat ein CD-Wechsel stattgefunden? -------------------------------}
Const IOCTLI_MEDIA_CHANGED = 9;
Type MSCDEX_MedChng = Record
       bControlBlockCode,                                      { = 9 }
       bMedChng          : Byte;         { nach Aufruf: NOT_CHANGED, }
     End;                                       { DONT_KNOW, CHANGED }

Const
     NOT_CHANGED = 1;
     DONT_KNOW   = 0;
     CHANGED     = $FF;

{-- Ermittelt CD-Informationen --------------------------------------}
Const IOCTLI_AUDIO_DISK_INFO = 10;
Type MSCDEX_DiskInfo = Record
      bControlBlockCode,                                      { = 10 }
      bLowestTrack,                                   { erster Track }
      bHighestTrack     : Byte;                      { Letzter Track }
      bLeadOutTrack     : Longint;          { Gre der CD (REDBOOK) }
    End;

{-- Ermittelt Informationen ber einen bestimmten Track -------------}
Const IOCTLI_AUDIO_TRACK_INFO = 11;
Type MSCDEX_TnoInfo=record
       bControlBlockCode,                                     { = 11 }
       bTrackNo         : Byte;{ Nummer des zu untersuchenden Tracks }
       lStartingPoint   : Longint;       { Adresse im REDBOOK Format }
       bTrackControlInfo: Byte;          { Info ber Track           }
     End;

{- Definitionen fr TrackControlInfo --------------------------------}
Const
     TCI_4AUDIO_CHANNELS         = $80;
     TCI_DATA_TRACK              = $40;
     TCI_TRACK_MASK              = $C0;
     TCI_PRE_EMPHASIS            = $10;
     TCI_DIGITAL_COPY_PROHIBITED = $20;

{-- Ermittle aktuelle Track- und Disk-Spielzeit ---------------------}
Const IOCTLI_AUDIO_QUERY_CHANNEL = 12;
Type MSCDEX_QInfo = Record
       bControlBlockCode,                                     { = 12 }
       bCONTROLandADR,      { TrackControl, Adressierung (s.TnoInfo) }
       bTrackNo,               { Nummer des aktuellen Tracks (Lieds) }
       bIndex,           { Aktueller Index innerhalb aktuellem Track }
       bTrackMin,                         { Aktuelle Track-Spielzeit }
       bTrackSec,
       bTrackFrame,
       bZero,                                              { immer 0 }
       bDiskMin,                            { Gesamtspielzeit der CD }
       bDiskSec,
       bDiskFrame        : Byte;
     End;

{-- Ermittelt Sub-Channel-Info whrend des Abspielens ---------------}
Const IOCTLI_AUDIO_SUB_CHANNEL_INFO = 13;
Type MSCDEX_SubChanInfo = Record
       bControlBlockCode      : Byte;                         { = 13 }
       lStartingSector        : Longint;         { = REDBOOK-Adresse }
       lpTransferAdress       : Pointer;     { = Adresse des Buffers }
       lNumberOfSectorsToRead : Longint; { = Anzahl zu lesender Sub- }
    End;                                 { Channel-Informationen.    }

{-- Ermittle UPC/EAN-Nummer der eingelegten CD. ---------------------}
Const IOCTLI_UPCODE = 14;
type MSCDEX_UPCode = Record
       bControlBlockCode,                                     { = 14 }
       bCONTROLandADR   : Byte;                         { s. TnoInfo }
       bUPC_EAN         : array[0..6] of Byte;{ 13 Stellige BCD-Zahl }
       bZero,                          { (letztes Nibble unbebnutzt) }
       bAFrame          : Byte;
     End;

{-- Ermittelt Audio-Status ------------------------------------------}
Const IOCTLI_AUDIO_STATUS_INFO = 15;
Type MSCDEX_AudStat = Record
       bControlBlockCode : Byte;                              { = 15 }
       wAudioStatus      : Word; { Bit 0 = Audio Pause, 1-15 reserv. }
       lResumeStart,     { Startposition fr DEVCMD_RESUME (REDBOOK) }
       lResumeEnd        : Longint;        { Endposition fr         }
     End;                                  { DEVCMD_RESUME (REDBOOK) }

Const
  {-- INPUTFLUSH verwendet Standard-RequestHeader -------------------}
  DEVCMD_INPUTFLUSH  = 7;

  {-- OUTPUTFLUSH verwendet Standard-RequestHeader ------------------}
  DEVCMD_OUTPUTFLUSH = 11;

  {-- ber das WRITE-Kommando werden einige Laufwerks-Parameter -----}
  {-- verndert. Dazu wird der MSCDEX_IOCTLIO-Block verwendet (s.o.) }
  DEVCMD_IOCTL_WRITE = 12;

{-- ffnet CD-Lade --------------------------------------------------}
Const IOCTLO_EJECT_DISK = 0;
Type MSCDEX_Eject = Record
       bControlBlockCode : Byte;                               { = 0 }
     End;

{-- Verriegelt Tr --------------------------------------------------}
Const IOCTLO_LOCK_DOOR = 1;
Type MSCDEX_LockDoor = Record
       bControlBlockCode,                                      { = 1 }
       bLock            : Byte;{ = 1: Verriegeln, 0: ffnen der Lade }
     End;

{-- Zurcksetzen des Laufwerks --------------------------------------}
Const IOCTLO_RESET_DRIVE = 2;
Type MSCDEX_Reset = Record
       bControlBlockCode : Byte;                               { = 2 }
     End;

{-- Gegenstck zu IOCTLI_GET_AUDIO_CHANNEL_INFO ---------------------}
Const IOCTLO_SET_AUDIO_CHANNEL_INFO = 3;

{-- Gegenstck zu IOCTLI_READ_DRIVE_BYTES ---------------------------}
Const IOCTLO_WRITE_DRIVE_BYTES = 4;

{-- Laufwerkslappe schlieen ----------------------------------------}
Const IOCTLO_CLOSE_TRAY = 5;
Type MSCDEX_CloseTray = Record
       bControlBlockCode : Byte;                               { = 5 }
    End;

Const
 {-- Device ffnen --------------------------------------------------}
 DEVCMD_DEVICE_OPEN  = 13;

 {-- Device Schlieen -----------------------------------------------}
 DEVCMD_DEVICE_CLOSE = 14;

 {-- Liest Daten von einer CD ---------------------------------------}
 DEVCMD_READLONG     = 128;

{- Folgende Struktur wird beim Lesen & Schreiben von Daten verwendet }
Type MSCDEX_ReadWriteL = Record
       ReqHdr           : RequestHeader;
       bAdressingMode   : Byte;           { = HSG(0) oder REDBOOK(1) }
       lpTransferAdress : Pointer;           { = Adresse des Buffers }
       wNumSec          : Word;{ =Anz. der zu bertragenden Sektoren }
       lStartingSector  : Longint;         { = (Schreib)Leseposition }
       bDataReadWriteMode,                { = bertragungsmodus s.u. }
       bInterleaveSize,            { = Anzahl zus.hngender Sektoren }
       bInterleaveSkipFactor : Byte;             { = Gre der Lcke }
  End;

  {- bDataReadWriteMode-Definitionen: -------------------------------}
  { Beim Lesen werden nur RAW und COOKED verwendet:                  }
Const
   COOKED     = 0;
   RAW        = 1;
  {-- Beim Schreiben kommen folgende Konstanten zum Tragen: ---------}
   MODE0      = 0;      { Lschen der Sektoren (berschreiben mit 0) }
   MODE1      = 1;  { Beschreiben der Sektoren mit angegebenen Daten }
   MODE2FORM1 = 2;                               { COOKED Data-Write }
   MODE2FORM2 = 3;                                  { RAW Data-Write }

{- bermittelt den nchsten zu lesenden Sektor an Treiber, soda     }
{- dieser vorrausschauend den Sektor 'Cachen' kann.                  }
Const DEVCMD_READLONG_PREFETCH = 130;

{- Setzt Position des (Schreib)Lesekopfes auf gesamter CD -----------}
Const DEVCMD_SEEK = 131;
Type MSCDEX_SeekReq = Record
       ReqHdr           : RequestHeader;
       bAdressingMode   : Byte;           { = HSG(0) oder REDBOOK(1) }
       lpTransferAdress : Pointer; { wird ignoriert daher = 0 setzen }
       wNumSec          : Word;    { wird ignoriert daher = 0 setzen }
       lStartingSector  : Longint;         { = anzufahrende Position }
     End;

{ Audio-Track abspielen ---------------------------------------------}
Const DEVCMD_PLAY_AUDIO = 132;
Type MSCDEX_PlayReq = Record
       ReqHdr         : RequestHeader;
       bAdressingMode : Byte;             { = HSG(0) oder REDBOOK(1) }
       lStartSector,                 { = erster zu spielender Sektor }
       lSectorCount   : Longint;   { = Anzahl zu spielender Sektoren }
     End;

{- Audio-Ausgabe stoppen (verwendet Standard RequestHeader) ---------}
Const DEVCMD_STOP_AUDIO = 133;

{- Daten schreiben (s. READ_LONG) -----------------------------------}
Const DEVCMD_WRITELONG = 134;

{- geschriebene Daten mit denen im Speicher vergleichen -------------}
Const DEVCMD_WRITELONG_VERIFY = 135;

{- Durch einmaliges STOP_AUDIO unterbrochene Musik weiterspielen ----}
const DEVCMD_RESUME_AUDIO = 136;
Type HSG_DIR_ENTRY = Record
       len_dr,                               { Lnge dieser Struktur }
       XAR_len     : Byte;{XAR-Lnge in LBN (logische Block Nummern) }
       loc_extentI,                     { Start-LBN's (Intel Format) }
       loc_extentM,                  { Start-LBN's (Molorola Format) }
       data_lenI,                        { Dateilnge (Intel Format) }
       data_lenM   : Longint;         { Dateilnge (Motorola Format) }
       record_time : Array[1..5] of Byte;           { Datum und Zeit }
       file_flags_hsg,                                       { Flags }
       reserved,                                        { reserviert }
       il_size,      { Interleave: Anzahl zusammenhngender Sektoren }
       il_skip     : Byte;             { Interleave: Gre der Lcke }
       VSSNI,            { Volume Set Sequence Number (Intel Format) }
       VSSNM,         { Volume Set Sequence Number (Motorola Format) }
       len_fi      : Word;                         { Lnge des Namen }
{  BYTE  file_id[ ... ];Filename, variable Lnge (max. 32 Zeichen)   }
{  BYTE  padding; Pad-Byte (damit System-Daten Word-aligned sind     }
{  BYTE  sys_data[ ... ];          System Daten (variable Lnge)     }
End;

Type ISO_DIR_ENTRY = Record
    len_dr,                                  { Lnge dieser Struktur }
    XAR_len     : Byte;  { XAR-Lnge in LBN (logische Block Nummern) }
    loc_extentI,                          { Start-LBN (Intel Format) }
    loc_extentM,                       { Start-LBN (Motorola Format) }
    data_lenI,                           { Dateilnge (Intel Format) }
    data_lenM   : Longint;            { Dateilnge (Motorola Format) }
    record_time : array[0..6] of Byte;              { Datum und Zeit }
    file_flags_iso,                                          { Flags }
    il_size,         { Interleave: Anzahl zusammenhngender Sektoren }
    il_skip     : Byte;                { Interleave: Gre der Lcke }
    VSSNI,               { Volume Set Sequence Number (Intel Format) }
    VSSNM       : Word;{Volume Set Sequence Number (Motorola Format) }
    len_fi      : Byte;                            { Lnge des Namen }
{  BYTE  file_id[ ... ];Filename, variable Lnge (max. 32 Zeichen)   }
{  BYTE  padding; Pad-Byte (damit System-Daten Word-aligned sind     }
{  BYTE  sys_data[ ... ];          System Daten (variable Lnge)     }
  End;

Type DIR_ENTRY = Record
 XAR_len      : Byte;    { XAR-Lnge in LBN (logische Block Nummern) }
 loc_extent   : Longint;                                 { Start-LBN }
 lb_size      : Word ;   { Gre eines (LBN). Wenn LBN-Gre = 2048, }
                         { dann entspricht ein LBN einem physischen  }
                         { Sektor/Frame.                             }
 data_len     : Longint;                        { Dateilnge (BYTES) }
 record_time  : Array[0..6] of Byte;                { Datum und Zeit }
 file_flags,                                                 { Flags }
 il_size,            { Interleave: Anzahl zusammenhngender Sektoren }
 il_skip      : Byte;                  { Interleave: Gre der Lcke }
 VSSN         : Word;                   { Volume Set Sequence Number }
 len_fi       : Byte;                        { Lnge des Dateinamens }
 file_id      : Array[0..37] of Byte;{ Dateiname mit #0 abgschlossen }
 file_version : Word;                                 { File-Version }
 len_su       : Byte;                        { Lnge der Systemdaten }
 su_data      : Array[0..219] of byte;                 { Systemdaten }
End;
                                   { Klartext fr Device-Status-Bits }
Const
 DevStat : Array[0..12,0..1] of String =
           ( ( 'Door is open', 'Door is closed' ),
             ( 'Door is unlocked', 'Door is locked' ),
             ( 'Drive reads COOKED and RAW', 'Drive reads only COOKED' ),
             ( 'Drive reads and writes', 'Drive only reads' ),
             ( 'Drive reads DATA and plays AUDIO/VIDEO tracks',
               'Drive reads only DATA tracks' ),
             ( 'Drive supports ISO 9600 interleave',
               'Drive supports no interleave'),
             ( 'reserved', 'reserved' ),
             ( 'Drive supports prefetching requests',
               'Drive doesn''t support prefetching requests'),
             ( 'Drive supports audio-channel manipulation',
               'Drive doesn''t support audio-channel manipulation' ),
             ( 'Drive supports HSG and REDBOOK adressing',
               'Drive supports HSG Adressing only'),
             ( 'reserved', 'reserved' ),
             ( 'No disk in drive', 'Disk in drive'),
             ( 'Drive supports R/W sub-channel-info',
               'Drive doesn''t support R/W sub-channel-info' ) );

Function cd_IsError( wStatus : Word ) : Boolean;

Function cd_GetReqHdrError( var ReqHdr : RequestHeader ) : Integer;

Function cd_IsReqHdrBusy( var ReqHdr : RequestHeader ) : Boolean;

Function cd_IsReqHdrDone( var ReqHdr : RequestHeader ) : Boolean;

Function cd_GetDeviceHeaderAdress( iCD_Drive_Letter : Integer;
                                   var RA           : MSCDEX_Raddr) : Integer;

Function cd_GetLocationOfHead( iCD_Drive_Letter : Integer;
                               var LH           : MSCDEX_LocHead ) : Integer;

Function cd_GetAudioChannelInfo( iCD_Drive_Letter : Integer;
                                 var AI           : MSCDEX_AudInfo ) : Integer;

Function cd_ReadDriveBytes( iCD_Drive_Letter : Integer;
                            var DB           : MSCDEX_DrvBytes ) : Integer;

Function cd_GetDevStat( iCD_Drive_Letter : Integer;
                        var DS           : MSCDEX_DevStat ) : Integer;

Procedure cd_PrintDevStat( var DS : MSCDEX_DevStat );

Function cd_GetSectorSize( iCD_Drive_Letter : Integer;
                           var SEC          : MSCDEX_SectSize;
                           iReadMode        : Integer ) : Integer;

Function cd_GetVolumeSize( iCD_Drive_Letter : Integer;
                           var VS           : MSCDEX_VolSize ) : Integer;

Function cd_GetMediaChanged( iCD_Drive_Letter : Integer;
                             var MC           : MSCDEX_MedChng ) : Integer;

Function cd_GetAudioDiskInfo( iCD_Drive_Letter : Integer;
                              var DI           : MSCDEX_DiskInfo ) : Integer;

Function cd_GetAudioTrackInfo( iCD_Drive_Letter : integer;
                               iTrackNo         : Integer;
                               var TI           : MSCDEX_TnoInfo ) : Integer;

Function cd_IsDataTrack( var TI : MSCDEX_TnoInfo  ) : Boolean;

Function cd_GetTrackLen( iCD_Drive_Letter, iTno : Integer ) : Longint;

Function cd_QueryAudioChannel( iCD_Drive_Letter : Integer;
                               var QI           : MSCDEX_QInfo  ) : Integer;

Function cd_GetAudioSubChannel( iCD_Drive_Letter : Integer;
                                var SCI          : MSCDEX_SubChanInfo ):Integer;

Function cd_GetUPCode( iCD_Drive_Letter : Integer;
                       var UPC          : MSCDEX_UPCode ) : Integer;

Procedure cd_PrintUPCode( var UPC : MSCDEX_UPCode );

Function cd_GetAudioStatusInfo( iCD_Drive_Letter : Integer;
                                var AS           : MSCDEX_AudStat ) : Integer;

Function cd_Eject( iCD_Drive_Letter : Integer ) : Integer;

Function cd_LockDoor( iCD_Drive_Letter : Integer; iLock : Boolean ) : Integer;

Function cd_ResetDrive( iCD_Drive_Letter : Integer ) : Integer;

Function cd_SetAudioChannelInfo( iCD_Drive_Letter : Integer;
                                 var AI           : MSCDEX_AudInfo  ) : Integer;

Function cd_WriteDriveBytes( iCD_Drive_Letter : Integer;
                             var DB           : MSCDEX_DrvBytes ) : Integer;

Function cd_CloseTray( iCD_Drive_Letter : Integer ) : Integer;

Function cd_PlayAudio( iCD_Drive_Letter : integer;
                       iAdressMode      : Integer;
                       lStart           : longint;
                       lLen             : Longint ):Integer;

Function cd_StopAudio( iCD_Drive_Letter : Integer ) : Integer;

Function cd_ResumeAudio( iCD_Drive_Letter : Integer ) : Integer;

Function cd_Seek( iCD_Drive_Letter : integer;
                  iAdressingMode   : Integer;
                  lSector          : Longint ) : Integer;

Function cd_ReadLong( iCD_Drive_Letter : integer;
                      iAdressMode      : Integer;
                      lStart           : Longint;
                      iNumSec          : Integer;
                      lpReadData       : Pointer;
                      iDataMode        : Integer ) : Integer;

Function cd_ReadLongPrefetch( iCD_Drive_Letter : integer;
                              iAdressMode      : Integer;
                              lStart           : Longint ) : Integer;

Function cd_WriteLong( iCD_Drive_Letter : integer;
                       iAdressMode      : Integer;
                       lStart           : Longint;
                       iNumSec          : Integer;
                       lpWriteData      : Pointer;
                       iDataMode        : Integer ) : Integer;

Function cd_WriteLongVerify( iCD_Drive_Letter : integer;
                             iAdressMode      : Integer;
                             lStart           : Longint;
                             iNumSec          : Integer;
                             lpVerifyData     : Pointer;
                             iDataMode        : Integer ) : Integer;

Procedure cd_PrintDiskTracks( iCD_Drive_Letter : Integer );

Procedure cd_FastForward( iCD_Drive_Letter : Integer );

Function cd_PrintActPlay( iCD_Drive_Letter : Integer ) : Boolean;

Procedure cd_PrintSektor( lpSektor : Pointer;
                          iSize    : integer;
                          iCols    : integer;
                          iRows    : Integer );

Procedure cd_PrintDirEntry( var DirEntry : DIR_ENTRY  );
Function BCDOut( z:Integer ) : String;
Function IntOut( z:Integer; l:Integer ) : String;

Implementation

{********************************************************************}
{ cd_IsError : Beschreibt Status-Wort einen Fehler?                  }
{*------------------------------------------------------------------*}
{ Eingabe : wStatus - zu testendes Status-Wort                       }
{ Ausgabe : TRUE  - Status-Wort beschreibt Fehler                    }
{           FALSE - Status-Wort beschreibt keinen Fehler             }
{*------------------------------------------------------------------*}
{ Info    : Diese Funktion stellt fest, ob das Fehler-Bit (Bit15)    }
{           gesetzt ist.                                             }
{********************************************************************}
Function cd_IsError( wStatus : Word ) : Boolean;

Begin
  if ( wStatus and DEV_ERROR ) <> 0 then cd_IsError := TRUE
                                    else cd_IsError := FALSE;
End;

{********************************************************************}
{ cd_GetReqHdrError : Ermittle Fehler der letzten Device-            }
{                     Kommunikation                                  }
{*------------------------------------------------------------------*}
{ Eingabe : ReqHdr - Request-Header, der auf Fehler untersucht       }
{                    werden soll.                                    }
{ Ausgabe :  > 0 - FehlerCode + 1                                    }
{           == 0 - Kein Fehler                                       }
{********************************************************************}
Function cd_GetReqHdrError( var ReqHdr : RequestHeader  ) : Integer;

Begin
  if cd_IsError( ReqHdr.wStatus ) then             { Bit 15 testen }
    cd_GetReqHdrError := ( ReqHdr.wStatus and $00FF ) + 1
  else
    cd_GetReqHdrError := 0;
End;

{********************************************************************}
{ cd_IsReqHdrBusy : Ist Device noch mit Ausfhrung von Kommandos     }
{                   beschftigt ?                                    }
{*------------------------------------------------------------------*}
{ Eingabe : ReqHdr - Device-Header                                   }
{ Ausgabe : TRUE  - Device noch mit alten Kommandos beschftigt      }
{           FALSE - Device nicht mehr beschftigt                    }
{********************************************************************}
Function cd_IsReqHdrBusy( var ReqHdr : RequestHeader  ) : Boolean;

Begin
  if ( ReqHdr.wStatus and DEV_BUSY ) <> 0 then
    cd_IsReqHdrBusy := TRUE
  else
    cd_IsReqHdrBusy := FALSE;    { Bit 9 }
End;

{********************************************************************}
{ cd_IsReqHdrDone : Ist letztes Device-Kommando vollstndig          }
{                   ausgefhrt worden ?                              }
{*------------------------------------------------------------------*}
{ Eingabe : ReqHdr - Device-Header                                   }
{ Ausgabe : TRUE  - Device-Kommando wurde ausgefhrt                 }
{           FALSE - Device noch BUSY                                 }
{********************************************************************}
Function cd_IsReqHdrDone( var ReqHdr : RequestHeader ) : Boolean;

Begin
  if( ReqHdr.wStatus and DEV_DONE ) <> 0 then
    cd_IsReqHdrDone := TRUE
  else
    cd_IsReqHdrDone := FALSE;   { Bit 8 }
End;

{- Wrapper-Funktionen fr MSCDEX-Device Kommandos -------------------}

{********************************************************************}
{ cd_GetDeviceHeaderAdress : Ermittle Adresse des Device-Headers     }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           RA               - Raddr-ControlBlock                    }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : lpRA->lpDeviceHeaderAdress enthlt nach Aufruf die          }
{        Adresse des Headers des fr das angegebene CD-Laufwerk      }
{        zustndigen Device-Treibers                                 }
{********************************************************************}
Function cd_GetDeviceHeaderAdress( iCD_Drive_Letter : Integer;
                                   var RA           : MSCDEX_Raddr ) : Integer;
Begin
  RA.bControlBlockCode := IOCTLI_GET_DEVHDR_ADRESS;          { = 0 }
  cd_GetDeviceHeaderAdress :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @RA,
                         sizeof( MSCDEX_Raddr ) );
End;

{********************************************************************}
{ cd_GetLocationOfHead : Ermittle Position des (Schreib-)/Lesekopfes }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           LH               - LocHead-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : LH.bAdressingMode mu vor dem Aufruf Adressierungsmodus     }
{        enthalten (HSG(0) oder REDBOOK(1)). Nach dem Aufruf befin-  }
{        det sich dann in LH.lLocation die aktuelle Position des     }
{        Schreib-/Lesekopfes.                                        }
{********************************************************************}
Function cd_GetLocationOfHead( iCD_Drive_Letter : Integer;
                               var LH           : MSCDEX_LocHead ) : Integer;
Begin
  LH.bControlBlockCode := IOCTLI_GET_HEAD_LOCATION;          { = 1 }
  cd_GetLocationOfHead :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @LH,
                         sizeof( MSCDEX_Raddr ) );
End;

{********************************************************************}
{ cd_GetAudioChannelInfo : Ermittle Zuordnung der CD-Kanle zu       }
{                          CD-Ausgngen                              }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           AI               - AudInfo-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : AI.Output0-4 und AI.Volume0-4 enthalten Zuordnung und Laut- }
{        strke der 4 mglichen Kanle                               }
{********************************************************************}
Function cd_GetAudioChannelInfo( iCD_Drive_Letter : Integer;
                                 var AI           : MSCDEX_AudInfo ) : Integer;
Begin
  AI.bControlBlockCode := IOCTLI_GET_AUDIO_CHANNEL_INFO;     { = 4 }
  cd_GetAudioChannelInfo :=
     MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                          DEVCMD_IOCTL_READ,
                          @AI,
                          sizeof( MSCDEX_AudInfo ) );
End;

{********************************************************************}
{ cd_ReadDriveBytes : Lies Laufwerk-Spezifische Daten                }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           DB               - DrvBytes-ControlBlock                 }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : DB.bNumReadBytes enthlt vor dem Aufruf die Anzahl der      }
{        zu lesenden Bytes. Nach dem Aufruf stehen in DB.bReadBuffer }
{        die gelesenen Bytes. Sollen mehr als 128 Bytes gelesen      }
{        werden, werden die vorherigen 128Byte-Blcke verworfen,     }
{        um den durch INT(NumReadBytes/128) bestimmten Block zu      }
{        bergeben.                                                  }
{********************************************************************}
Function cd_ReadDriveBytes( iCD_Drive_Letter : Integer;
                            var DB           : MSCDEX_DrvBytes ) : Integer;
Begin
  DB.bControlBlockCode := IOCTLI_READ_DRIVE_BYTES;           { = 5 }
  cd_ReadDriveBytes :=
     MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                          DEVCMD_IOCTL_READ,
                          @DB,
                          sizeof( MSCDEX_DrvBytes ) );
End;

{********************************************************************}
{ cd_GetDevStat : Lese Device-Status (Laufwerkszustand )             }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           DS               - DevStat-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : DS.lDeviceStatus enthlt nach Aufruf den Device-Status      }
{********************************************************************}
Function cd_GetDevStat( iCD_Drive_Letter : Integer;
                        var DS           : MSCDEX_DevStat ) : Integer;
Begin
  DS.bControlBlockCode := IOCTLI_GET_DEV_STAT;               { = 6 }
  cd_GetDevStat :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @DS,
                         sizeof( MSCDEX_DevStat ) );
End;

{********************************************************************}
{ cd_PrintDevStat : Device-Status ausgeben                           }
{*------------------------------------------------------------------*}
{ Eingabe : DS - zuvor mittels GetDevStat ermittelter Device-Status  }
{********************************************************************}
Procedure cd_PrintDevStat( var DS : MSCDEX_DevStat );
var i : Integer;
Begin
     { Alle Bits durchgehen und Meldung fr gesetztes/gelschtes Bit }
     { ausgeben                                                      }
  for i := 0 to 12 do
    if (DS.lDeviceStatus and ( 1 shl i ) ) <> 0 then
     writeln( DevStat[i,0] ) else writeln( DevStat[i,1] );
End;

{********************************************************************}
{ cd_GetSectorSize : Ermittle Gre eines Sektors                    }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           Sec              - SectSize-ControlBlock                 }
{           iReadMode        - COOKED ( 0 ) oder RAW ( 1 )           }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : lpSec->wSectorSize enthlt nach Aufruf Sektor-Gre (Bytes) }
{********************************************************************}
Function cd_GetSectorSize( iCD_Drive_Letter : Integer;
                           var SEC          : MSCDEX_SectSize;
                           iReadMode        : Integer ) : Integer;
Begin
  SEC.bControlBlockCode := IOCTLI_SECTOR_SIZE;               { = 7 }
  SEC.bReadMode := Byte (iReadMode);
  cd_GetSectorSize :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @SEC,
                         sizeof( MSCDEX_SectSize ) );
End;

{********************************************************************}
{ cd_GetVolumeSize : Ermittle Anzahl der Sektoren einer CD           }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           VS               - VolSize-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : VS.lVolumeSize enthlt nach Aufruf Anzahl der Sektoren      }
{        der eingelegten CD.                                         }
{********************************************************************}
function cd_GetVolumeSize( iCD_Drive_Letter : Integer;
                           var VS           : MSCDEX_VolSize ) : Integer;
Begin
  VS.bControlBlockCode := IOCTLI_VOLUME_SIZE;                { = 8 }
  cd_GetVolumeSize :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @VS,
                         sizeof( MSCDEX_VolSize ) );
End;

{********************************************************************}
{ cd_GetMediaChanged : Wurde CD gewechselt?                          }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           MC               - MedChng-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : MC.bMediaChanged enthlt nach Aufruf Status des Mediums     }
{        1   : CD seit letzem Aufruf nicht gewechselt.               }
{        0   : vielleicht gewechselt                                 }
{        255 : CD wurde seit letztem Aufruf gewechselt.              }
{********************************************************************}
Function cd_GetMediaChanged( iCD_Drive_Letter : Integer;
                             var MC           : MSCDEX_MedChng ) : Integer;
Begin
  MC.bControlBlockCode := IOCTLI_MEDIA_CHANGED;              { = 9 }
  cd_GetMediaChanged :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @MC,
                         sizeof( MSCDEX_MedChng ) );
End;

{********************************************************************}
{ cd_GetAudioDiskInfo : Ermittle Anzahl der Titel (Stcke, Tracks)   }
{                       einer Audio-CD.                              }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           DI               - DiskInfo-ControlBlock                 }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : DI.bLowestTrack, DI.bHighestTrack enthalten nach Aufruf die }
{        Nummer des ersten bzw. letzten Liedes.                      }
{        DI.bLeadOutTrack enthlt Adresse des ersten unbenutzen      }
{        Sektors (nach letzem Titel)                                 }
{********************************************************************}
function cd_GetAudioDiskInfo( iCD_Drive_Letter : Integer;
                              var DI           : MSCDEX_DiskInfo ) : Integer;
Begin
  DI.bControlBlockCode := IOCTLI_AUDIO_DISK_INFO;           { = 10 }
  cd_GetAudioDiskInfo :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @DI,
                         sizeof( MSCDEX_DiskInfo ) );
End;

{********************************************************************}
{ cd_GetAudioTrackInfo : Ermittle Daten zu einem Titel               }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iTrackNo         - Nummer des Tracks, fr den Daten      }
{                              ermittelt werden sollen.              }
{           TI               - TackInfo-ControlBlock                 }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : TI.StartingPoINT     - REDBOOK Adresse des Titelstarts      }
{        TI.bTrackControlInfo - Informationen ber Track-Typ         }
{********************************************************************}
function cd_GetAudioTrackInfo( iCD_Drive_Letter : integer;
                               iTrackNo         : Integer;
                               var TI           : MSCDEX_TnoInfo  ) : Integer;
Begin
  TI.bControlBlockCode := IOCTLI_AUDIO_TRACK_INFO;          { = 11 }
  TI.bTrackNo :=  Byte (iTrackNo);
  cd_GetAudioTrackInfo :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @TI,
                         sizeof( MSCDEX_TnoInfo ) );
End;

{********************************************************************}
{ cd_IsDataTrack : Ist ein Track ein reiner Daten Track, oder        }
{                  enthlt er Audio-Informationen?                   }
{*------------------------------------------------------------------*}
{ Eingabe : TI - TrackInfo-ControlBlock                              }
{ Ausgabe : TRUE  - TI beschreibt Daten-Track                        }
{           FALSE - Track enthlt Audio-Daten                        }
{********************************************************************}
function cd_IsDataTrack( var TI : MSCDEX_TnoInfo  ) : Boolean;
Begin
  if ( ( TI.bTrackControlInfo and TCI_TRACK_MASK ) = TCI_DATA_TRACK )
    then cd_IsDataTrack := TRUE
    else cd_IsDataTrack := FALSE;
End;

{********************************************************************}
{ cd_GetTrackLen : Ermittle Gre eines Tracks (Anzahl Sektoren bzw.  }
{                  Frames)                                           }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iTrackNo         - Nummer des Tracks, fr den Daten      }
{                              ermittelt werden sollen.              }
{ Ausgabe : Anzahl der Frames des angegebenen Tracks                 }
{*------------------------------------------------------------------*}
{ Info : Beim letzten Track mu der Lead-Out-Track als Terminator    }
{        verwendet werden.                                           }
{********************************************************************}
Function cd_GetTrackLen( iCD_Drive_Letter, iTno : Integer ) : Longint;
var DI     : MSCDEX_DiskInfo;
    TI     : MSCDEX_TnoInfo;
    VS     : MSCDEX_VolSize;
    lStart,
    lEnd   : Longint;

Begin
  cd_GetAudioDiskInfo( iCD_Drive_Letter, DI );
  if( ( iTno < DI.bLowestTrack ) or ( iTno > DI.bHighestTrack ) ) then
    cd_GetTrackLen:=0                     { kein gltiger Track }
  else
    Begin
      cd_GetAudioTrackInfo( iCD_Drive_Letter, iTno, TI );
      lStart := REDBOOK2HSG( TI.lStartingPoint );
      if( iTno < DI.bHighestTrack ) then
        Begin                        { Differenz aufeinanderfolg. Tracks }
         cd_GetAudioTrackInfo( iCD_Drive_Letter, iTno + 1, TI );
         lEnd := REDBOOK2HSG( TI.lStartingPoint );
        End
      else          { Differenz [Ende der CD - letzter Track] }
        Begin
          cd_GetVolumeSize( iCD_Drive_Letter, VS );
          lEnd := VS.lVolumeSize - 150;          { Frames in HSG umrechnen }
        End;
      cd_GetTrackLen := lEnd - lStart;
    End;
End;

{********************************************************************}
{ cd_QueryAudioChannel : Ermittle Online-Informationen ber zur Zeit }
{                        gespielten Audio-Titel                      }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           QI               - QInfo-ControlBlock                    }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : QI.bTrackNo/bIndex/bTrackMin/bTrackSec/bTrackFrame:         }
{        Informationen ber aktuell abgespieltes Lied (Nummer und    }
{        Spielzeit)                                                  }
{        QI.bDiskMin/bDiskSec/bDiskFrame: bisherige Gesammtspiezeit  }
{                                                                    }
{        Diese Funktion kann laufend aufgerufen werden, um die       }
{        aktuellen Spielzeiten zu erhalten.                          }
{********************************************************************}
Function cd_QueryAudioChannel( iCD_Drive_Letter : Integer;
                               var QI           : MSCDEX_QInfo ) : Integer;
Begin
  QI.bControlBlockCode := IOCTLI_AUDIO_QUERY_CHANNEL;       { = 12 }
  cd_QueryAudioChannel :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @QI,
                         sizeof( MSCDEX_QInfo ) );
End;

{********************************************************************}
{ cd_GetAudioSubChannelInfo : Ermittle Sub-Channel-Informationen     }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           SCI              - SubChanInfo-ControlBlock              }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info :    In der in SCI.lpTransferAdress angegebenen Adresse       }
{           finden sich nach Aufruf dieser Funktion die              }
{           Sub-Channel-Inforationen der angebenenen Sektoren        }
{           (SCI.lStartingSektor, SCI.lNumberOfSectorsToRead)        }
{           Fr jeden Sektor mu ein 96Bytes groer Buffer bereit-   }
{           gestellt werden.                                         }
{********************************************************************}
function cd_GetAudioSubChannel( iCD_Drive_Letter : Integer;
                                var SCI          : MSCDEX_SubChanInfo ) : Integer;
Begin
  SCI.bControlBlockCode := IOCTLI_AUDIO_SUB_CHANNEL_INFO;   { = 13 }
  cd_GetAudioSubChannel :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @SCI,
                         sizeof( MSCDEX_SubChanInfo ) );
End;

{********************************************************************}
{ cd_GetUPCode : Ermittle Universal-Product-Code einer CD            }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           UPC              - UPCode-ControlBlock                   }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info :    Nach dem Aufruf befindet sich in UPC.bUPC_EAN eine       }
{           13 stellige BCD-Zahl, die den UP-Code der aktuellen CD   }
{           angibt.                                                  }
{********************************************************************}
Function cd_GetUPCode( iCD_Drive_Letter : Integer;
                       var UPC          : MSCDEX_UPCode ) : Integer;
Begin
  UPC.bControlBlockCode := IOCTLI_UPCODE;                   { = 14 }
  cd_GetUPCode :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @UPC,
                         sizeof( MSCDEX_UPCode ) );
End;

{********************************************************************}
{ cd_PrintUPCode : Gib UPC BCD-Zahl aus.                             }
{*------------------------------------------------------------------*}
{ Eingabe : UPC - UPCode-ControlBlock                                }
{*------------------------------------------------------------------*}
{ Info : Letztes Nibble des bUPC_EAN-Arrays wird nicht verwendet     }
{********************************************************************}
procedure cd_PrintUPCode( var UPC : MSCDEX_UPCode  );
var i : Integer;
Begin
  Write(' UPC: ');
  for i := 0 to 5 do                         { die ersten 12 Ziffern }
    Write( BCDOut( UPC.bUPC_EAN[ i ]  ) );
  Inc(i);                        { letzte Ziffer "von Hand" ausgeben }
  Writeln( char ( ( ( UPC.bUPC_EAN[ i ] shr  4 ) and $0F ) +
           ord( '0' ) ) );
End;

{********************************************************************}
{ cd_GetAudioStatusInfo : Ermittle aktuellen Audio Zustand           }
{                         (Play/Pause)                               }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           AS               - AudStat-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info :    Nach dem Aufruf gibt Bit 0 der Variablen                 }
{           AS.wAudioStatus Auskunft ber den Spielmodus des CD-     }
{           Laufwerks. (Gesetzt = Pause)                             }
{           AS.lResumeStart enthlt die als nchstes abzuspielende   }
{           Position bei einem RESUME. AS.lResumeEnd gibt an, wann   }
{           das Abspielen des Liedes gestoppt werden soll.           }
{********************************************************************}
Function cd_GetAudioStatusInfo( iCD_Drive_Letter : Integer;
                                var AS           : MSCDEX_AudStat ) : Integer;
Begin
  AS.bControlBlockCode := IOCTLI_AUDIO_STATUS_INFO;         { = 15 }
  cd_GetAudioStatusInfo :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_READ,
                         @AS,
                         sizeof( MSCDEX_AudStat ) );
End;

{********************************************************************}
{ cd_Eject : Lade ffnen (ausfahren)                                 }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Vor dem ffnen der Lade wird diese entriegelt               }
{********************************************************************}
Function cd_Eject( iCD_Drive_Letter : Integer ) : Integer;

var E : MSCDEX_Eject ;

Begin
  E.bControlBlockCode := IOCTLO_EJECT_DISK;                     { = 0 }
  cd_Eject := MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                                   DEVCMD_IOCTL_WRITE,
                                   @E,
                                   sizeof( MSCDEX_Eject ) );
End;

{********************************************************************}
{ cd_LockDoor : Lade ver-/entriegeln                                 }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iLock            - == 0 : Lade entriegeln                }
{                              <> 0 : Lade verriegeln                }
{ Ausgabe : Device-Status                                            }
{********************************************************************}
function cd_LockDoor( iCD_Drive_Letter : Integer; iLock : Boolean ) : Integer;

var LD : MSCDEX_LockDoor;

Begin
  LD.bControlBlockCode := IOCTLO_LOCK_DOOR;                     { = 1 }
  if iLock then LD.bLock := 1 else LD.bLock := 0;
  cd_LockDoor := MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                                      DEVCMD_IOCTL_WRITE,
                                      @LD,
                                      sizeof( MSCDEX_LockDoor ) );
End;

{********************************************************************}
{ cd_Reset : Laufwerk zurcksetzen                                   }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : Device-Status                                            }
{********************************************************************}
function cd_ResetDrive( iCD_Drive_Letter : Integer ) : Integer;

var R : MSCDEX_Reset;

Begin
  R.bControlBlockCode := IOCTLO_RESET_DRIVE;                  { = 2 }
  cd_ResetDrive := MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                                        DEVCMD_IOCTL_WRITE,
                                        @R,
                                        sizeof( MSCDEX_Reset ) );
End;

{********************************************************************}
{ cd_SetAudioChannelInfo : Setze Zuordnung der CD-Kanle zu          }
{                          CD-Ausgngen                              }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           AI               - AudInfo-ControlBlock                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : AI.Output0-4 und AI.Volume0-4 enthalten Zuordnung und Laut- }
{        strke der 4 mglichen Kanle                               }
{********************************************************************}
Function cd_SetAudioChannelInfo( iCD_Drive_Letter : Integer;
                                 var AI           : MSCDEX_AudInfo ) : Integer;

Begin
  AI.bControlBlockCode := IOCTLO_SET_AUDIO_CHANNEL_INFO;     { = 3 }
  cd_SetAudioChannelInfo :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_WRITE,
                         @AI,
                         sizeof( MSCDEX_AudInfo ) );
End;

{********************************************************************}
{ cd_WriteDriveBytes : bergibt Laufwerk-Spezifische Daten           }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           DB               - DrvBytes-ControlBlock                 }
{ Ausgabe : Device-Status                                            }
{********************************************************************}
Function cd_WriteDriveBytes( iCD_Drive_Letter : Integer;
                             var DB           : MSCDEX_DrvBytes ) : Integer;

Begin
  DB.bControlBlockCode := IOCTLO_WRITE_DRIVE_BYTES;           { = 4 }
  cd_WriteDriveBytes :=
    MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                         DEVCMD_IOCTL_WRITE,
                         @DB,
                         sizeof( MSCDEX_DrvBytes ) );
End;

{********************************************************************}
{ cd_CloseTray : Lade verschliessen (einfahren)                      }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter : Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : Device-Status                                            }
{********************************************************************}
function cd_CloseTray( iCD_Drive_Letter : Integer ) : Integer;

var CT : MSCDEX_CloseTray;

Begin
  CT.bControlBlockCode := IOCTLO_CLOSE_TRAY;                    { = 5 }
  cd_CloseTray := MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                                       DEVCMD_IOCTL_WRITE,
                                       @CT,
                                       sizeof( MSCDEX_CloseTray ) );
End;

{********************************************************************}
{ cd_PlayAudio : Spielt Musikdaten ab                                }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressMode      - HSG(0) oder REDBOOK(1)-Adressen?      }
{           lStart           - Adresse ab der mit dem Abspielen      }
{                              begonnen werden soll                  }
{           lLen             - Anzahl der abzuspielenden Sektoren    }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info    : Durch dieses Kommando werden BUSY und PAUSE-Bit (Audio-  }
{           Stat) aufdatiert. Ist der CD-Player schon mit dem Ab-    }
{           spielen von Musikdaten beschftigt, mu vor einem neuen  }
{           PLAY_AUDIO-Request erst ein STOP_AUDIO-Request           }
{           abgesendet werden. Untersttzt das Laufwerk kein Audio   }
{           wird dieser Befehl ignoriert.                            }
{********************************************************************}
Function cd_PlayAudio( iCD_Drive_Letter : Integer;
                       iAdressMode      : Integer;
                       lStart           : longint;
                       lLen             : Longint ) : Integer;

var PR : MSCDEX_PlayReq;

Begin
  if lLen < 0 then lLen := 0;                       { Immer Positiv }
  PR.ReqHdr.bLength  := sizeof( RequestHeader );
  PR.ReqHdr.bSubUnit := 0;
  PR.ReqHdr.bCommand := Byte ( DEVCMD_PLAY_AUDIO );
  PR.bAdressingMode  := Byte ( iAdressMode );
  PR.lStartSector    := lStart;
  PR.lSectorCount    := lLen;
  cd_PlayAudio := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, PR.ReqHdr );
End;

{********************************************************************}
{ cd_StopAudio : Stoppt Musikausgabe                                 }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info    : Durch das Play-Kommando abzuspielende Musikdaten werden  }
{           solange abgespielt, bis alle zu spielenden Sektoren aus- }
{           gegegeben wurden. Um die Musikausgabe vorzeitig zu       }
{           unterbrechen, mu das Stop-Kommando gesendet werden.     }
{           Wird ein Play-Request durch einen einmaligen Stop-Aufruf }
{           unterbrochen, kann dieser durch ein Resume wieder        }
{           gestartet werden. Es wird ab der letzten Position        }
{           weitergespielt. Nach zwei aufeinanderfolgenden Stop-     }
{           Requests kann die Audioausgabe nur durch einen neuen     }
{           Play-Request gestartet werden.                           }
{********************************************************************}
Function cd_StopAudio( iCD_Drive_Letter : Integer ) : Integer;

var ReqHdr : RequestHeader;

Begin
  ReqHdr.bLength  := sizeof( RequestHeader );
  ReqHdr.bSubUnit := 0;
  ReqHdr.bCommand := Byte ( DEVCMD_STOP_AUDIO );

  cd_StopAudio := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, ReqHdr );
End;

{********************************************************************}
{ cd_ResumeAudio : Startet einmalig unterbrochene Musikausgabe       }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : Device-Status                                            }
{********************************************************************}
Function cd_ResumeAudio( iCD_Drive_Letter : Integer ) : Integer;

var ReqHdr : RequestHeader ;

Begin
  ReqHdr.bLength  := sizeof( RequestHeader );
  ReqHdr.bSubUnit := 0;
  ReqHdr.bCommand := Byte ( DEVCMD_RESUME_AUDIO );

  cd_ResumeAudio := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, ReqHdr );
End;

{********************************************************************}
{ cd_Seek : Position (Schreib-)/Lesekopf                             }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressingMode   - HSG(0) oder REDBOOK(1)                }
{           lSector          - Anzufahrender Sektor                  }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Die Positionierung des (Schreib-)/Lesekopf unterscheidet    }
{        sich von der Positionierung des Dateizeiger innerhalb einer }
{        Datei. Beim CD-Laufwerk kann Seek Dateibergreifend (auch   }
{        bei der Audio-Ausgabe) verwendet werden! Bei der Audio-     }
{        Ausgabe kann ein Seek allerdings zur Positionierung ber    }
{        das Ende der Ausgabe (Anzahl abzuspielender Sektoren)       }
{        bewegt werden.                                              }
{********************************************************************}
Function cd_Seek( iCD_Drive_Letter : integer;
                  iAdressingMode   : Integer;
                  lSector          : Longint ) : Integer;

var SR : MSCDEX_SeekReq;

Begin
  SR.ReqHdr.bLength   := sizeof( RequestHeader );
  SR.ReqHdr.bSubUnit  := 0;
  SR.ReqHdr.bCommand  := Byte ( DEVCMD_SEEK );
  SR.bAdressingMode   := Byte ( iAdressingMode );
  SR.lpTransferAdress := NIL;
  SR.wNumSec          := 0;
  SR.lStartingSector  := lSector;
  cd_Seek := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, SR.ReqHdr );
End;

{********************************************************************}
{ cd_ReadLong : Liest Anzahl von Sektoren (s. AbsoluteDiskRead)      }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressingMode   - HSG(0) oder REDBOOK(1)                }
{           lStart           - Erster zu lesender Sektor             }
{           iNumSec          - Anzahl der zu lesenden Sektoren       }
{           lpReadData       - Adresse des Lesebuffers               }
{           iDataMode        - RAW(1) oder COOKED(0)                 }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Ein RAW-Sektor ist 2352 Bytes gro. COOKED-Sektoren         }
{        hingegen nur 2048 Bytes.                                    }
{********************************************************************}
Function cd_ReadLong( iCD_Drive_Letter : integer;
                      iAdressMode      : Integer;
                      lStart           : Longint;
                      iNumSec          : Integer;
                      lpReadData       : Pointer;
                      iDataMode        : Integer ) : Integer;

var RWL : MSCDEX_ReadWriteL;

Begin
  RWL.ReqHdr.bLength        := sizeof( RequestHeader );
  RWL.ReqHdr.bSubUnit       := 0;
  RWL.ReqHdr.bCommand       := Byte ( DEVCMD_READLONG );
  RWL.bAdressingMode        := Byte ( iAdressMode );
  RWL.lpTransferAdress      := lpReadData;
  RWL.wNumSec               := iNumSec;
  RWL.lStartingSector       := lStart;
  RWL.bDataReadWriteMode    := Byte ( iDataMode );
  RWL.bInterleaveSize       := 1;
  RWL.bInterleaveSkipFactor := 1;

  cd_ReadLong := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, RWL.ReqHdr );
End;

{********************************************************************}
{ cd_ReadLongPrefetch : Teilt Laufwerk den nchsten zu lesenden      }
{                       Sektor mit.                                  }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressingMode   - HSG(0) oder REDBOOK(1)                }
{           lStart           - Erster zu lesender Sektor             }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Um die langsame Geschwindigkeit der CD's ein wenig auszu-   }
{        gleichen, kann mittels Prefetching ein vereinfachtes Cache  }
{        System (ein Sektor gro) aufgebaut werden. Dieser Befehl    }
{        bermittelt dem Device-Treiber dazu den nchsten zu         }
{        lesenden Sektor, der im 'vorbeigehen' gelesen wird.         }
{********************************************************************}
Function cd_ReadLongPrefetch( iCD_Drive_Letter : integer;
                              iAdressMode      : Integer;
                              lStart           : Longint ) : Integer;
var RWL : MSCDEX_ReadWriteL;
Begin
  RWL.ReqHdr.bLength        := sizeof( RequestHeader );
  RWL.ReqHdr.bSubUnit       := 0;
  RWL.ReqHdr.bCommand       := Byte ( DEVCMD_READLONG_PREFETCH );
  RWL.bAdressingMode        := Byte ( iAdressMode );
  RWL.lpTransferAdress      := NIL;
  RWL.wNumSec               := 1;
  RWL.lStartingSector       := lStart;
  RWL.bDataReadWriteMode    := 0;
  RWL.bInterleaveSize       := 0;
  RWL.bInterleaveSkipFactor := 0;

  cd_ReadLongPrefetch := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, RWL.ReqHdr );
End;

{********************************************************************}
{ cd_WriteLong : Schreibt Anzahl von Sektoren (s. AbsoluteDiskWrite) }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressingMode   - HSG(0) oder REDBOOK(1)                }
{           lStart           - Erster zu schreibender Sektor         }
{           iNumSec          - Anzahl der zu schreibenden Sektoren   }
{           lpWriteData      - Adresse des Schreibbuffers            }
{           iDataMode        - Mode0(0), Mode1(1),                   }
{                              Mode2Form1(2), Mode2Form2(3)          }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Mode0 -      Sektoren werden mit Nullen beschriben          }
{        Mode1 -      COOKED-Write (2048 Byte groe Sektoren)        }
{        Mode2Form1 - COOKED-Write (2048 Byte groe Sektoren)        }
{        Mode2Form2 - RAW-Write (2352 Byte groe Sektoren)           }
{********************************************************************}
Function cd_WriteLong( iCD_Drive_Letter : integer;
                       iAdressMode      : Integer;
                       lStart           : Longint;
                       iNumSec          : Integer;
                       lpWriteData      : Pointer;
                       iDataMode        : Integer ) : Integer;

var RWL : MSCDEX_ReadWriteL;

Begin
  RWL.ReqHdr.bLength        := sizeof( RequestHeader );
  RWL.ReqHdr.bSubUnit       := 0;
  RWL.ReqHdr.bCommand       := Byte ( DEVCMD_WRITELONG );
  RWL.bAdressingMode        := Byte ( iAdressMode );
  RWL.lpTransferAdress      := lpWriteData;
  RWL.wNumSec               := iNumSec;
  RWL.lStartingSector       := lStart;
  RWL.bDataReadWriteMode    := Byte ( iDataMode );
  RWL.bInterleaveSize       := 0;
  RWL.bInterleaveSkipFactor := 0;

  cd_WriteLong := MSCDEX_SendDeviceRequest( iCD_Drive_Letter, RWL.ReqHdr );
End;

{********************************************************************}
{ cd_WriteLongVerify : Schreibt und prft Daten auf CD mit Daten     }
{                      im Speicher                                   }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{           iAdressingMode   - HSG(0) oder REDBOOK(1)                }
{           lStart           - Erster zu berprfenden Sektor        }
{           iNumSec          - Anzahl der zu berprfenden Sektoren  }
{           lpVerifyData     - Adresse des Prfbuffers               }
{           iDataMode        - Mode1(1),                             }
{                              Mode2Form1(2), Mode2Form2(3)          }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info : Mode1:      COOKED-Verify (2048 Byte groe Sektoren)        }
{        Mode2Form1: COOKED-Verify (2048 Byte groe Sektoren)        }
{        Mode2Form2: RAW-Verify (2352 Byte groe Sektoren)           }
{********************************************************************}
Function cd_WriteLongVerify( iCD_Drive_Letter : integer;
                             iAdressMode      : Integer;
                             lStart           : Longint;
                             iNumSec          : Integer;
                             lpVerifyData     : Pointer;
                             iDataMode        : Integer ) : Integer;

var RWL : MSCDEX_ReadWriteL;

Begin
  RWL.ReqHdr.bLength        := sizeof( RequestHeader );
  RWL.ReqHdr.bSubUnit       := 0;
  RWL.ReqHdr.bCommand       := Byte ( DEVCMD_WRITELONG_VERIFY );
  RWL.bAdressingMode        := Byte ( iAdressMode );
  RWL.lpTransferAdress      := lpVerifyData;
  RWL.wNumSec               := iNumSec;
  RWL.lStartingSector       := lStart;
  RWL.bDataReadWriteMode    := Byte ( iDataMode );
  RWL.bInterleaveSize       := 0;
  RWL.bInterleaveSkipFactor := 0;

  cd_WriteLongVerify := MSCDEX_SendDeviceRequest( iCD_Drive_Letter,
                                                  RWL.ReqHdr );
End;

{********************************************************************}
{ IntOut : Integer inkl. fhrender Nullen ausgeben                   }
{*------------------------------------------------------------------*}
{ Eingabe : z : Auszugebende Zahl                                    }
{           l : Lnge des Ausgabestrings inkl. Nullen                }
{ Ausgabe : Formatierter Zahlenstring                                }
{********************************************************************}
function IntOut( z:Integer; l:Integer ):String;
var dummy:String;
begin
     Str( z, dummy );
     while( length( dummy ) < l ) do  dummy := '0' + dummy;
     IntOut := dummy;
end;

{********************************************************************}
{ BCDOut : BCD-Zahl in String wandeln                                }
{*------------------------------------------------------------------*}
{ Eingabe : z : Auszugebende Zahl                                    }
{ Ausgabe : Formatierter Zahlenstring                                }
{********************************************************************}
function BCDOut( z:Integer ):String;
var dummy:String;
begin
  dummy := char ( ( ( z shr  4 ) and $0F ) + ord( '0' ) ) +
           char ( ( ( z and $0F ) + ord( '0' ) ) );
  BCDOut := dummy;
end;

{********************************************************************}
{ cd_PrintDiskTracks : Zeige alle Titel und zugehrige Spielzeiten   }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{********************************************************************}
Procedure cd_PrintDiskTracks( iCD_Drive_Letter : Integer );

var DI     : MSCDEX_DiskInfo;
    TI     : MSCDEX_TnoInfo;
    i,
    iMin,
    iSec,
    iFrame : Integer;
    lLen   : Longint;

Begin

  if ( not cd_IsError( cd_GetAudioDiskInfo( iCD_Drive_Letter, DI ) ) ) then
    for i := DI.bLowestTrack to DI.bHighestTrack do
      Begin
        if ( not cd_IsError(cd_GetAudioTrackInfo(
             iCD_Drive_Letter,i, TI ) ) ) then
          Begin
            REDBOOK2Time( TI.lStartingPoint, iMin, iSec, iFrame );
            Write( ' ', IntOut( i, 2 ), ': ',
                        IntOut( iMin, 2 ), ':',
                        IntOut( iSec, 2 ), '.',
                        IntOut( iFrame,2 ), '   ' );
            lLen := cd_GetTrackLen( iCD_Drive_Letter, i );
            Frame2Time( lLen, iMin, iSec, iFrame );
            Write( '  Lnge: ', IntOut( i, 2 ), ': ',
                                IntOut( iMin, 2 ), ':',
                                IntOut( iSec, 2 ), '.',
                                IntOut( iFrame, 2 ) );
            Write( ' Typ :' );
            case ( TI.bTrackControlInfo and TCI_TRACK_MASK ) of
              TCI_DATA_TRACK: Write( 'Data ' );
              TCI_4AUDIO_CHANNELS: Write( 'Audio/Quadro ' )
              else Write('Audio/Stereo ');
            End;
            if ( TI.bTrackControlInfo and TCI_PRE_EMPHASIS ) <> 0 then
              Write('Pre-Emphasis ');
            if ( TI.bTrackControlInfo and TCI_DIGITAL_COPY_PROHIBITED ) <> 0 then
              Write('Digital Copy prohibited ');
            Writeln;
          End;
      End;
End;

{********************************************************************}
{ cd_FastForward : Aktuelle Musikausgabe 2 Sekunden vorlaufen lassen }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{********************************************************************}
Procedure cd_FastForward( iCD_Drive_Letter : Integer );

var AS   : MSCDEX_AudStat;
    lHSG : Longint;

Begin
  cd_GetAudioStatusInfo( iCD_Drive_Letter, AS );
  cd_StopAudio( iCD_Drive_Letter );
  lHSG := AS.lResumeStart + Time2Frame( 0, 2, 0 );
  cd_PlayAudio( iCD_Drive_Letter,
                REDBOOK,
                HSG2REDBOOK( lHSG ),
                REDBOOK2HSG( AS.lResumeEnd ) - lHSG );
End;

{********************************************************************}
{ cd_PrintActPlay : Zeige Spielzeiten und Titel das aktuell abge-    }
{                   spielten Titel an.                               }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung( 0=A, etc )          }
{ Ausgabe : TRUE             - CD-Player noch mit Audio-Ausgabe be-  }
{                              schftigt                             }
{           FALSE            - CD-Player mit abspielen fertig        }
{********************************************************************}
Function cd_PrintActPlay( iCD_Drive_Letter : Integer ) : Boolean;

var iStat : Integer;
    QI    : MSCDEX_QInfo;

Begin
  iStat := cd_QueryAudioChannel( iCD_Drive_Letter, QI );
  Write( ' Track: ', BCDOut( QI.bTrackNo ), ' Index: ',
                     IntOut( QI.bIndex, 2),'  ' );
  Write( ' Track: ', IntOut( QI.bTrackMin, 2 ), ':',
                     IntOut( QI.bTrackSec, 2 ), '.',
                     IntOut( QI.bTrackFrame, 2 ) );
  Write( ' Disk: ', IntOut( QI.bDiskMin, 2), ':',
                    IntOut( QI.bDiskSec, 2 ), '.',
                    IntOut( QI.bDiskFrame, 2) ,#13 );
  if ( iStat and DEV_BUSY ) <> 0 then
    cd_PrintActPlay := TRUE else cd_PrintActPlay := FALSE;
End;

{********************************************************************}
{ Hex : wandelt einen Hex-Wert in einen Hex-String um                }
{*------------------------------------------------------------------*}
{ Eingabe : hz  : Word mit Hex-Zahl                                  }
{ Ausgabe : Der Hex-String                                           }
{********************************************************************}
Function Hex( hz : Word ) : String;

var i  : Integer;
    hs : String;

Begin
 hs:='';
 Repeat
  i := hz mod 16;
  hz := hz div 16;
  case i of
   0..9: hs := chr( i + ord( '0' ) ) + hs;
   10..15: hs := chr( i - 10 + ord( 'A' ) ) + hs;
  end;
 Until hz = 0;
 if length( hs ) mod 2 <> 0 then hs := '0' + hs;
 hex := hs;
End;

{********************************************************************}
{ cd_PrintSektor : Stelle Inhalt eines Sektors in Zeichen / Hex dar. }
{*------------------------------------------------------------------*}
{ Eingabe : lpSektor - Adresse der Sektor-Daten                      }
{           iSize    - Gre des Sektors                             }
{           iCols    - Anzahl darzustellender Spalten                }
{           iRows    - Anzahl anzuzeigender Zeilen, nach denen auf   }
{                      Tastendruck gewartet werden soll.             }
{                      0 - Nicht warten                              }
{********************************************************************}
Procedure cd_PrintSektor( lpSektor : Pointer;
                          iSize    : integer;
                          iCols    : integer;
                          iRows    : Integer );
var lpS     : ^byte;
    i, j, k : Integer;

Begin
  lpS := lpSektor;
  k := 0;
  i := 0;
  while i < iSize do
  Begin
    for j := 0 to iCols-1 do
      if ( i + j < iSize )  then
        Begin
          lps := Pointer ( Longint ( lpSektor ) + i + j );
          if ( ( lpS^ >= $20 ) and ( lpS^ <= $7f ) ) then
            Write( chr( lpS^ ) ) else Write( '.' )
        End
      else
        Write( ' ' );

    Write( '  ' );
    for j := 0 to iCols-1 do
      if( i + j < iSize ) then
        begin
          lps := Pointer ( Longint ( lpSektor ) + i + j );
          Write( Hex( lpS^ ) );
        end
      else
        Write( '   ' );

    Writeln;
    inc( i, iCols);
    if ( iRows > 0 ) then inc( k );
    if ( ( k = iRows ) and ( i < iSize ) ) then
    Begin
      Write( '< Taste >'#13 );
      ReadKey;
      k := 0;
    End;
  End;
End;

{********************************************************************}
{ cd_PrintDirEntry : Zeige Elemente einer DirEntry-Struktur an       }
{*------------------------------------------------------------------*}
{ Eingabe : DirEntry - anzuzeigende Struktur                         }
{********************************************************************}
Procedure cd_PrintDirEntry( var DirEntry : DIR_ENTRY );

var i : Integer;

Begin
  Writeln(' XAR-Lnge                 : ', Integer ( DirEntry.XAR_len ) );
  Writeln(' Start-LBN                 : ', DirEntry.loc_extent );
  Writeln(' LBN-Gre                 : ', DirEntry.lb_size );
  Writeln(' Datei-Lnge               : ', DirEntry.data_len );
  Writeln(' Flags                     : ', DirEntry.file_flags );
  Writeln(' Interleave-Size           : ', Integer( DirEntry.il_size ) );
  Writeln(' Interleave-Skip           : ', Integer(DirEntry.il_skip ) );
  Writeln(' Volume Set Sequence Number: ', DirEntry.VSSN );
  Writeln(' Lnge des Dateinamen      : ', DirEntry.len_fi );
  Write(' DateiName                 : ' );
  i:=0;
  While ( ( DirEntry.file_id[i] <> 0 ) and ( i < 38 ) ) do
    Begin
      Write( char ( DirEntry.file_id[i] ) );
      Inc(i);
    End;
  Writeln;
  Writeln(' Version                   : ',
          DirEntry.file_version shr 8, '.',
          DirEntry.file_version and $00FF );
  Writeln(' Systemdaten Lnge         : ', Integer(DirEntry.len_su ) );
End;

End.


