{********************************************************************}
{                          C D U T I L . P A S                       }
{*------------------------------------------------------------------*}
{ Aufgabe        : Utils fr MSCDEX-Programmierung                   }
{                  Helper, Wrapper, Interfaces                       }
{*------------------------------------------------------------------*}
{ Autor          : Michael Tischer / Bruno Jennrich                  }
{ entwickelt am  : 08.04.1994                                        }
{ letztes Update : 10.10.1994                                        }
{********************************************************************}

{$A-}                           { kein Word-Alignment von Strukturen }

Unit CDUTIL;

Interface

Const
 HSG     = 0;
 REDBOOK = 1;

{- MSCDEX - Funktionscodes ( bergeben in AL ) ----------------------}
 MSCDEX_GET_NO_OF_DRIVE_LETTERS     =   $00;
 MSCDEX_GET_DRIVE_DEVICE_LIST       =   $01;

Type
DeviceHeader = Record { Kopf eines DOS-Gertetreibers }
  lpNextHdr         : Pointer;{ Adresse des nchsten Device-Treibers }
  wDevAttributes,                                { Treiber-Attribute }
  wStrategy,                           { Offset der Strategy-Routine }
  wInterrupt        : Word;           { Offset der Interrupt-Routine }
  bName             : Array[0..7] of Byte;       { Name des Treibers }
  wReserved         : Word;
  bDriveLetter,                                 { Laufwerksbuchstabe }
  bNumberOfSubUnits : Byte;              { Anzahl der Untereinheiten }
end;

DevHeadPtr = ^DeviceHeader; { Zeiger auf einen Treiberkopf }

DevElement = Record
  bSubUnit       : Byte;                        { Nummer der SubUnit }
  lpDeviceHeader : DevHeadPtr;       { Adresse des Device-Headers }
End;

{- Request-Header ---------------------------------------------------}
RequestHeader = Record
  bLength,                              { Lnge der Request-Struktur }
  bSubUnit,                                { Nummer der Untereinheit }
  bCommand  : Byte;                        { auszufhrendes Kommando }
  wStatus   : Word;                                   { Fehlerstatus }
  bReserved : Array[0..7] of Byte;         { werden je nach Kommando }
End;                                                   { umdefiniert }

{ ber das IOCTL_READ-Kommando werden vielfltige Statusinfor- }
{ mationen ber Treiber, Laufwerk und eingelegte CD abgefragt. }
MSCDEX_IOCTLIO = Record
  ReqHdr            : RequestHeader;
  bMediaDescriptor  : Byte;                                { immer 0 }
  lpTransferAdress  : Pointer;        { = Adresse des Control-Blocks }
  wTransferCount,                       { = Gre des Control-Blocks }
  wStartingSectorNo : Word;                                { immer 0 }
  lpVolumeID        : Pointer;                           { immer NIL }
End;

Const
 {MSCDEX - Funktionen }
 MSCDEX_GET_COPYRIGHT_FILENAME      =   $02;
 MSCDEX_GET_ABSTRACT_FILENAME       =   $03;
 MSCDEX_GET_BIB_DOC_FILENAME        =   $04;
 MSCDEX_READ_VTOC                   =   $05;
 MSCDEX_DEBUG_ON                    =   $06;
 MSCDEX_DEBUG_OFF                   =   $07;
 MSCDEX_ABSOLUTE_DISK_READ          =   $08;
 MSCDEX_ABSOLUTE_DISK_WRITE         =   $09;
 MSCDEX_RESERVED                    =   $0A;

 {MSCDEX - Funktionen ab V2.00 }
 MSCDEX200_CD_ROM_DRIVE_CHECK       =   $0B; 
 MSCDEX200_GET_VERSION              =   $0C;
 MSCDEX200_GET_DRIVE_LETTERS        =   $0D;
 MSCDEX200_GETSET_VD_PREFERENCE     =   $0E;
 MSCDEX200_GET_DIRECTORY_ENTRY      =   $0F;
 DIRECT_COPY = FALSE;
 STRUCT_COPY = TRUE;

 { MSCDEX - Funktionen ab V2.10 }
 MSCDEX210_SEND_DEVICE_REQUEST      =   $10;

 { Nummer des Multiplex-Interrupts }
 DOS_MULTIPLEX_INT                  =  $2F;

 { Erkennungscode der MSCDEX-MUX-Aufrufe }
 MSCDEX_MULTIPLEX_CODE = ($1500);

 iVersion  : Integer = 0;
 iNumUnits : Integer = -1;

Var

Strategy      : Pointer;                        { Funktionsadressen }
Interruptproc : Pointer;
S, O,                                             { Segment, Offset }
i,
iStart        : Integer;
                                      
cCDLetters    : Array[0..25 ] of byte; { Alle CD-Laufwerksbuchstaben }
DevElements   : Array[0..25 ] of DevElement;

Function Time2Frame( iMin, iSec, iFrame : Integer ) : Longint;

Function Time2HSG( iMin, iSec, iFrame : Integer) : Longint;

Function Time2REDBOOK( iMin, iSec, iFrame : Integer ) : Longint;

Procedure Frame2Time(     lFrame          : Longint;
                      var Min, Sec, Frame : Integer );

Procedure REDBOOK2Time(    lREDBOOK         : Longint;
                        var Min, Sec, Frame : Integer );

Procedure HSG2Time(     lHSG            : Longint;
                    var Min, Sec, Frame : Integer );

Function REDBOOK2HSG( lREDBOOK : Longint ) : Longint;

Function HSG2REDBOOK( lHSG:Longint ) : Longint;

Function MSCDEX_GetNumberOfDriveLetters( var Number,
                                             StartLetter : Integer ) : Boolean;

Function MSCDEX_Installed : Boolean;

Function MSCDEX_GetCDRomDriveDeviceList( DevElArray : pointer ) : Boolean;

Function MSCDEX_GetCopyrightFilename( iCD_Drive_Letter : Integer;
                                      lpName : Pointer ) : Boolean;

Function MSCDEX_GetAbstractFilename( iCD_Drive_Letter : Integer;
                                     lpName : Pointer ) : Boolean;

Function MSCDEX_GetBibDocFilename( iCD_Drive_Letter : Integer;
                               lpName : Pointer ) : Boolean;

Function MSCDEX_ReadVTOC( iCD_Drive_Letter : Integer;
                          lpSector : Pointer;
                          iNumSec : Integer ) : Integer;

Function MSCDEX_DebugOn : Boolean;

Function MSCDEX_DebugOff : Boolean;

Function MSCDEX_AbsoluteRead( iCD_Drive_Letter : Integer;
                              lpSector : Pointer;
                              iSecCnt : Integer;
                              lSecStart : Longint ) : Boolean;

Function MSCDEX_AbsoluteWrite( iCD_Drive_Letter : Integer;
                               lpSector : Pointer;
                               iSecCnt : Integer;
                               lSecStart : Longint ) : Boolean;

Function MSCDEX200_CDRomDriveCheck( iCD_Drive_Letter : Integer ) : Boolean;

Function MSCDEX200_GetVersion : Integer;

Function MSCDEX200_GetCDRomDriveLetters( lpLetters : Pointer ) : Boolean;

Function MSCDEX200_GetVDPreference(     iCD_Drive_Letter : Integer;
                                    var Preference       : Integer ) : Boolean;

Function MSCDEX200_SetVDPreference( iCD_Drive_Letter,
                                    iPreference : Integer ) : Boolean;

Function MSCDEX200_GetDirectoryEntry(     iCD_Drive_Letter : Integer;
                                          iCopyFlag        : Boolean;
                                          PathName         : String;
                                          lpData           : Pointer;
                                      var HSG              : Integer ) : Boolean;

Function MSCDEX210_SendDeviceRequest(     iCD_Drive_Letter : Integer;
                                      var ReqHdr           : RequestHeader ) : Integer;

Function _CallStrategy(     iCD_Drive_Letter : Integer;
                        var ReqHdr           : RequestHeader ) : Integer;

Function MSCDEX_SendDeviceRequest(     iCD_Drive_Letter : Integer;
                                   var ReqHdr           : RequestHeader ) : Integer;

Function MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                              iCommand : Integer;
                              lpControlBlock : Pointer;
                              iControlBlockSize : Integer ) : Integer;

Implementation
Uses DOS;

{********************************************************************}
{ Time2Frame : Wandelt Zeit in Anzahl der Frames (75 Frames = 1 sec) }
{*------------------------------------------------------------------*}
{ Eingabe : iMin   - Minute ( 0 - 59+ )                              }
{           iSec   - Sekunde ( 0 - 59 )                              }
{           iFrame - Frame ( 0 - 75 )                                }
{ Ausgabe : Anzahl der Frames                                        }
{********************************************************************}
Function Time2Frame( iMin, iSec, iFrame : Integer ) : Longint;

Begin
  Time2Frame := iMin * LongInt( 60 * 75 ) +
                iSec * LongInt( 75 ) +
                iFrame;
End;

{********************************************************************}
{ Time2HSG : Wandelt Zeit in HSG-Adresse                             }
{*------------------------------------------------------------------*}
{ Eingabe : iMin   - Minute ( 0 - 59+ )                              }
{           iSec   - Sekunde ( 0 - 59 )                              }
{           iFrame - Frame ( 0 - 75 )                                }
{ Ausgabe : Gewandelte HSG-Adresse                                   }
{********************************************************************}
Function Time2HSG( iMin, iSec, iFrame : Integer) : Longint;

Begin
  Time2HSG := Time2Frame( iMin, iSec, iFrame ) - 150;
End;

{********************************************************************}
{ Time2REDBOOK : Wandelt Zeit in REDBOOK-Adresse                     }
{*------------------------------------------------------------------*}
{ Eingabe : iMin   - Minute ( 0 - 59+ )                              }
{           iSec   - Sekunde ( 0 - 59 )                              }
{           iFrame - Frame ( 0 - 75 )                                }
{ Ausgabe : Gewandelte REDBOOK-Adresse                               }
{********************************************************************}
Function Time2REDBOOK( iMin, iSec, iFrame : Integer ) : Longint;

Begin
  Time2REDBOOK := ( ( Longint( iMin )   shl 16 ) or
                    ( Longint ( iSec )   shl  8 ) or
                    ( Longint ( iFrame ) shl  0 ) );
End;

{********************************************************************}
{ FrameTime : Splittet Anzahl von Frames (Sektoren) in Minute,       }
{             Sekunde, Frame auf.                                    }
{*------------------------------------------------------------------*}
{ Eingabe : lFrames  - zu wandelnde Anzahl von Frames (Sektoren)     }
{           Min      - Integer-Variable zur Aufnahme der Minuten     }
{                      Minuten                                       }
{           Sec      - dito fr Sekunden                             }
{           Frame    - dito fr Frames                               }
{********************************************************************}
Procedure Frame2Time(     lFrame          : Longint;
                      var Min, Sec, Frame : Integer );
Begin
  Min := Integer ( lFrame div ( LongInt( 60 * 75 ) ) );
  lFrame := lframe - Min * LongInt( 60 * 75 );
  Sec := Integer ( lFrame div LongInt( 75 ) );
  lFrame := lFrame - Sec * LongInt( 75 );
  Frame := Integer ( lFrame );
End;

{********************************************************************}
{ REDBOOK2Time : Splittet REDBOOK-Adresse in Minute, Sekunde, Frame  }
{                auf.                                                }
{*------------------------------------------------------------------*}
{ Eingabe : lREDBOOK - zu wandelnde REDBOOK-Adresse                  }
{           Min      - Integer-Variable zur Aufnahme der Minuten     }
{           Sec      - dito fr Sekunden                             }
{           Frame    - dito fr Frames                               }
{********************************************************************}
Procedure REDBOOK2Time(     lREDBOOK        : Longint;
                        var Min, Sec, Frame : Integer );
Begin
  Frame := Integer ( lREDBOOK shr  0 ) and $00FF;
  Sec   := Integer ( lREDBOOK shr  8 ) and $00FF;
  Min   := Integer ( lREDBOOK shr 16 ) and $00FF;
End;

{********************************************************************}
{ HSG2Time : Splittet HSG-Adresse in Minute, Sekunde, Frame auf.     }
{*------------------------------------------------------------------*}
{ Eingabe : lHSG  - zu wandelnde HSG-Adresse                         }
{           Min   - Integer-Variable zur Aufnahme der Minuten        }
{           Sec   - dito fr Sekunden                                }
{           Frame - dito fr Frames                                  }
{********************************************************************}
Procedure HSG2Time(     lHSG            : Longint;
                    var Min, Sec, Frame : Integer );
Begin
  Frame2Time( lHSG + LongInt( 150 ), Min, Sec, Frame );
End;

{********************************************************************}
{ REDBOOK2HSG : Wandle REDBOOK-Adresse in HSG-Adresse                }
{*------------------------------------------------------------------*}
{ Eingabe : lREDBOOK - zu wandelnde REDBOOK-Adresse                  }
{ Ausgabe : gewandelte HSG-Adresse                                   }
{********************************************************************}
Function REDBOOK2HSG( lREDBOOK : Longint ) : Longint;

var iMin, iSec, iFrame : Integer;

Begin
  REDBOOK2Time( lREDBOOK, iMin, iSec, iFrame );
  REDBOOK2HSG := Time2HSG( iMin, iSec, iFrame );
End;

{********************************************************************}
{ HSG2REDBOOK : Wandle HSG-Adresse in REDBOOK-Adresse                }
{*------------------------------------------------------------------*}
{ Eingabe : lHSG - zu wandelnde HSG-Adresse                          }
{ Ausgabe : gewandelte REDBOOK-Adresse                               }
{********************************************************************}
Function HSG2REDBOOK( lHSG : Longint ):Longint;

var iMin, iSec, iFrame : Integer;

Begin
  HSG2Time( lHSG, iMin, iSec, iFrame );
  HSG2REDBOOK := Time2REDBOOK( iMin, iSec, iFrame );
End;

{- MSCDEX-Interface Funktionen --------------------------------------}

{********************************************************************}
{ MSCDEX_GetNumberOfDriveLetters : Ermittle Anzahl der untersttzen  }
{                                  CD-Rom Laufwerke.                 }
{*------------------------------------------------------------------*}
{ Eingabe : Number      - Integer-Variable, die die Anzahl der Lauf- }
{                         werke aufnehmen soll (Ausgabeparameter)    }
{           StartLetter - ASCII-Code des ersten Laufwerksbuch-       }
{                         stabens der von einem CD-ROM Laufwerk      }
{                         verwendet wird.                            }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen                         }
{*------------------------------------------------------------------*}
{ Info :  Diese Funktion kann verwendet werden, um herauszufinden    }
{         ob berhaupt CD-ROM Laufwerke untersttzt werden bzw. ob   }
{         MSCDEX installiert wurde. Vor dem Aufruf wird BX auf 0     }
{         gesetzt. Ist dieses Register nach dem Aurfuf immer noch    }
{         gleich null, so wird kein CD-Rom Laufwerk untersttzt.     }
{********************************************************************}
Function MSCDEX_GetNumberOfDriveLetters( var Number,
                                             StartLetter : Integer ) : Boolean;
var regs:Registers;

Begin                                           { AX = $1500, BX = 0 }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_GET_NO_OF_DRIVE_LETTERS;
  regs.bx := 0;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then             { Carry-Flag gesetzt? }
  Begin
    Number := regs.bx;
    StartLetter := regs.cx;
    MSCDEX_GetNumberOfDriveLetters := TRUE;
  End
  else Begin
    Number := 0;
    StartLetter := 0;
    MSCDEX_GetNumberOfDriveLetters := FALSE;
  End;
End;

{********************************************************************}
{ MSCDEX_Installed : Ist MSCDEX berhaupt installiert ?              }
{*------------------------------------------------------------------*}
{ Info :  Diese Funktion testet, ob die Anzahl der untersttzen      }
{         Laufwerke von 0 verschieden ist.                           }
{********************************************************************}
Function MSCDEX_Installed : Boolean;
var iNumber, iStartLetter : Integer;
    dummy                 : boolean;

Begin
  iNumber := 0;
  dummy := MSCDEX_GetNumberOfDriveLetters( iNumber, iStartLetter );
  if iNumber = 0 then MSCDEX_Installed := FALSE
                 else MSCDEX_Installed := TRUE;
End;

{********************************************************************}
{ MSCDEX_GetCDRomDriveDeviceList : Ermittle Informationen ber       }
{                                  angeschlossene CD-Rom Laufwerke.  }
{*------------------------------------------------------------------*}
{ Eingabe : DevElArray - Zeiger auf Array mit Elementen vom Typ      }
{                        DevElement, in das die Informationen ber   }
{                        die angeschlossenen Laufwerke eingetragen   }
{                        werden.                                     }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info :  Ein DevElement enthlt nach dem Aufruf dieser Funktion die }
{         Adresse des Device-Headers zu der angegebenen SubUnit. Das }
{         dieser Funktion bergebene Array mu Platz fr alle ange-  }
{         schlossenen (bzw. untersttzten) CD-Rom Laufwerke ent-     }
{         halten. Die sicherste Methode besteht darin, Platz fr 26  }
{         Elemente (Laufwerksbuchstaben von A: bis Z:) zur Verfgung }
{         zu stellen und auf die ersten Elemente zuzugreifen.        }
{********************************************************************}
Function MSCDEX_GetCDRomDriveDeviceList( DevElArray : pointer ) : Boolean;

var regs : Registers;

Begin                                                   { AX = $1501 }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_GET_DRIVE_DEVICE_LIST ;
  regs.es := Word ( Longint ( DevElArray ) shr 16 );
  regs.bx := Word ( Longint ( DevElArray ) and $FFFF );
  intr( DOS_MULTIPLEX_INT, regs);
  if ( regs.flags and 1 ) = 0 then MSCDEX_GetCDRomDriveDeviceList := TRUE
                              else MSCDEX_GetCDRomDriveDeviceList := FALSE;
End;

{********************************************************************}
{ MSCDEX_GetCopyrightFilename : Ermittle Namen des Copyright-Files   }
{                               einer CD.                            }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpName           - Adresse eines 38 Zeichen groen       }
{                              Buffers zur Aufnahme des Namens.      }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info: Obwohl 38 Bytes zur Aufnahme des Filenamens zur Verfgung    }
{       stehen, werden nur die ersten 11 (8.3) Zeichen verwendet.    }
{********************************************************************}
Function MSCDEX_GetCopyrightFilename( iCD_Drive_Letter : Integer;
                                      lpName           : Pointer ) : Boolean;

var regs : Registers;

Begin                          { AX = $1502, CX = Laufwerksbuchstabe }
                                        { ES:BX = Adresse des Namens }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_GET_COPYRIGHT_FILENAME;
  regs.es := Longint ( lpName ) shr 16;
  regs.bx := Longint ( lpName ) and $ffff;
  regs.cx := iCD_Drive_Letter;
  intr( DOS_MULTIPLEX_INT, regs);
  if ( regs.flags and 1 ) = 0 then MSCDEX_GetCopyrightFilename := TRUE
                              else MSCDEX_GetCopyrightFilename := FALSE;
End;

{********************************************************************}
{ MSCDEX_GetAbstractFilename : Ermittle Namen des Abstract-Files     }
{                              einer CD.                             }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpName           - Adresse eines 38 Zeichen groen       }
{                              Buffers zur Aufnahme des Namens.      }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info: Obwohl 38 Bytes zur Aufnahme des Filenamens zur Verfgung    }
{       stehen, werden nur die ersten 11 (8.3) Zeichen verwendet.    }
{********************************************************************}
Function MSCDEX_GetAbstractFilename( iCD_Drive_Letter : Integer;
                                     lpName           : Pointer ) : Boolean;

var regs : Registers;

Begin                          { AX = $1503, CX = Laufwerksbuchstabe }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_GET_ABSTRACT_FILENAME;
  regs.es := Longint ( lpName ) shr 16;
  regs.bx := Longint ( lpName ) and $ffff;
  regs.cx := iCD_Drive_Letter;
  intr( DOS_MULTIPLEX_INT, regs);
  if ( regs.flags and 1 ) = 0 then MSCDEX_GetAbstractFilename := TRUE
                              else MSCDEX_GetAbstractFilename := FALSE;
End;

{********************************************************************}
{ MSCDEX_GetBibDocFilename : Ermittle Namen der bibliographischen    }
{                            Dokumentationsdatei einer CD.           }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpName           - Adresse eines 38 Zeichen groen       }
{                              Buffers zur Aufnahme des Namens.      }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info: Obwohl 38 Bytes zur Aufnahme des Filenamens zur Verfgung    }
{       stehen, werden nur die ersten 11 (8.3) Zeichen verwendet.    }
{********************************************************************}
Function MSCDEX_GetBibDocFilename( iCD_Drive_Letter : Integer;
                                   lpName           : Pointer ) : Boolean;
var regs : Registers;

Begin                          { AX = $1504, CX = Laufwerksbuchstabe }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_GET_BIB_DOC_FILENAME ;
  regs.es := Longint ( lpName ) shr 16;
  regs.bx := Longint ( lpName ) and $ffff;
  regs.cx := iCD_Drive_Letter;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then MSCDEX_GetBibDocFilename := TRUE
                              else MSCDEX_GetBibDocFilename := FALSE;
End;

{********************************************************************}
{ MSCDEX_ReadVTOC : Lese "Volume Table of Contents" (Inhalts-        }
{                   verzeichnis).                                    }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpSector         - Adresse eines 2048 Zeichen groen     }
{                              Buffers zur Aufnahme eines Sektors.   }
{           iNumSec          - Nummer des zu lesenden TOC-Sektors.   }
{ Ausgabe : <> 0 - Operation erfolgreich                             }
{                  $0101 - Gelesener Sektor = standard VTOC-Sektor   }
{                  $01FF - Gelesener Sektor = letzter VTOC-Sektor    }
{                  $0100 - Gelesener Sektor = anderer VTOC-Sektor    }
{           0    - Fehler (s. _dosernno )                            }
{********************************************************************}
Function MSCDEX_ReadVTOC( iCD_Drive_Letter : Integer;
                          lpSector         : Pointer;
                          iNumSec          : Integer ) : Integer;
var regs : Registers;
Begin                          { AX = $1505, CX = Laufwerksbuchstabe }
                                        { ES:BX = Adresse des Namens }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_READ_VTOC;
  regs.es := Longint ( lpSector ) shr 16;
  regs.bx := Longint ( lpSector ) and $ffff;
  regs.cx := iCD_Drive_Letter;
  regs.dx := iNumSec;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then MSCDEX_ReadVTOC := ( $0100 or regs.al )
                              else MSCDEX_ReadVTOC := 0;
End;

{********************************************************************}
{ MSCDEX_DebugOn : Debugging ermglichen.                            }
{                  (nur fr MSCDEX-Debug-Version)                    }
{*------------------------------------------------------------------*}
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX_DebugOn : Boolean;

var regs : Registers;

Begin                                                   { AX = $1506 }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_DEBUG_ON;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then  MSCDEX_DebugOn := TRUE
                              else  MSCDEX_DebugOn := FALSE;
End;

{********************************************************************}
{ MSCDEX_DebugOf : Debugging abschalten.                             }
{                  (nur fr MSCDEX-Debug-Version)                    }
{*------------------------------------------------------------------*}
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX_DebugOff : Boolean;

var regs : Registers;

Begin                                                   { AX = $1507 }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_DEBUG_OFF;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0  then MSCDEX_DebugOff := TRUE
                               else MSCDEX_DebugOff := FALSE;
End;

{********************************************************************}
{ MSCDEX_AbsoluteRead : Lese Daten aus Sektor(en)                    }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpSector         - Adresse der 2048 Zeichen groen       }
{                              Buffer zur Aufnahme der Sektoren.     }
{           iSecCnt          - Anzahl der zu lesenden Sektoren.      }
{           lSecStart        - Nummer des ersten zu lesenden Sektors }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info: Fr jeden zu lesenden Sektor mu ein 2048-Zeichen groer     }
{       Buffer bereitgestellt werden.                                }
{********************************************************************}
Function MSCDEX_AbsoluteRead( iCD_Drive_Letter : Integer;
                              lpSector         : Pointer;
                              iSecCnt          : Integer;
                              lSecStart        : Longint ) : Boolean;

var regs : Registers;

Begin                       { AX = $1508, CX = Laufwerkstbuchstabe   }
                            { DX = Anzahl zu lesender Sektoren       }
                            { ES:BX = Adresse des Lesebuffers        }
                            { SI:DI = Adresse des ersten Sektors     }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_ABSOLUTE_DISK_READ ;
  regs.es := Longint ( lpSector ) shr 16;
  regs.bx := Longint ( lpSector ) and $FFFF;
  regs.cx := iCD_Drive_Letter;
  regs.dx := iSecCnt;
  regs.si := lSecStart shr 16;
  regs.di := lSecStart and $ffff;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then  MSCDEX_AbsoluteRead := TRUE
                              else  MSCDEX_AbsoluteRead := FALSE;
End;

{********************************************************************}
{ MSCDEX_AbsoluteWrite: Schreibe Daten in Sektor(en)                 }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpSector  - Adresse der 2048 Zeichen groen Buffer,die   }
{                       die Daten enthalten.                         }
{           iSecCnt   - Anzahl der zu schreibenden Sektoren.         }
{           lSecStart - Nummer des ersten zu schreibenden Sektors    }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info: Fr jeden Sektor mu ein 2048-Zeichen groer Buffer bereit-  }
{       gestellt werden.                                             }
{********************************************************************}
Function MSCDEX_AbsoluteWrite( iCD_Drive_Letter : Integer;
                               lpSector : Pointer;
                               iSecCnt : Integer;
                               lSecStart : Longint ) : Boolean;

var regs : Registers;

Begin                       { AX = $1509, CX = Laufwerkstbuchstabe   }
                            { DX = Anzahl zu schreibender Sektoren   }
                            { ES:BX = Adresse des Schreibbuffers     }
                            { SI:DI = Adresse des ersten Sektors     }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX_ABSOLUTE_DISK_WRITE;
  regs.es := Longint ( lpSector ) shr 16;
  regs.bx := Longint ( lpSector ) and $ffff;
  regs.cx := iCD_Drive_Letter;
  regs.dx := iSecCnt;
  regs.si := lSecStart shr 16;
  regs.di := lSecStart and $ffff;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then MSCDEX_AbsoluteWrite := TRUE
                              else MSCDEX_AbsoluteWrite := FALSE;
End;

{********************************************************************}
{ MSCDEX200_CDRomDriveCheck : Ist Laufwerk ein CD-Rom Laufwerk?      }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX200_CDRomDriveCheck( iCD_Drive_Letter : Integer ) : Boolean;

var regs : Registers;

Begin                          { AX = $150B, CX = Laufwerksbuchstabe }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_CD_ROM_DRIVE_CHECK;
  regs.cx := iCD_Drive_Letter;
  regs.bx := 0;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( ( regs.bx = $ADAD ) and ( regs.ax <> 0 ) ) then
    MSCDEX200_CDRomDriveCheck := TRUE else
    MSCDEX200_CDRomDriveCheck := FALSE;
End;

{********************************************************************}
{ MSCDEX200_GetVersion : Ermittle MSCDEX-Version                     }
{*------------------------------------------------------------------*}
{ Ausgabe : Versionsnummer (Major.Minor)                             }
{*------------------------------------------------------------------*}
{ Info : Diese Funktion existiert erst seit MSCDEX V2.00! Kleinere   }
{        Versionsnummern werden als 1.00 angegeben!                  }
{********************************************************************}
Function MSCDEX200_GetVersion : Integer;

var regs : Registers;

Begin                                                  { AX = $150C }
  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_GET_VERSION;
  regs.bx := 0;                               { BX erst mal lschen }
  intr( DOS_MULTIPLEX_INT, regs );
  if ( ( regs.bx <> 0 ) and ( ( regs.flags and 1) = 0 ) ) then
    MSCDEX200_GetVersion:=regs.bx
  else
    if ( regs.flags and 1 ) = 0 then
      MSCDEX200_GetVersion:=$0100
    else
      MSCDEX200_GetVersion := -1;
                                              { V1.00 oder Fehler }
End;

{********************************************************************}
{ MSCDEX200_GetCDromDriveLetters : Ermittle von ALLEN CD-Laufwerken  }
{                                  verwendete Laufwerksbuchstaben.   }
{*------------------------------------------------------------------*}
{ Eingabe : lpLetters - Adress des Zeichen-Arrays, das Laufwerksbuch-}
{                       staben aufnehmen soll.                       }
{                       (s. GetNumberOfDriveLetters )                }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX200_GetCDRomDriveLetters( lpLetters : Pointer ) : Boolean;

var regs : Registers;

Begin                                                   { AX = $150D }
                                        { ES:BX = Adresse des Arrays }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_GET_DRIVE_LETTERS;
  regs.es := Longint ( lpLetters ) shr 16;
  regs.bx := Longint ( lpLetters ) and $ffff;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then MSCDEX200_GetCDRomDriveLetters := TRUE
                              else MSCDEX200_GetCDRomDriveLetters := FALSE;
End;

{********************************************************************}
{ MSCDEX200_GetVDPreference : Ermittle Volume Descriptor Preference  }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           lpiPreference    - Adresse des INT's, in das Preference  }
{                              gespeichert werden soll.              }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX200_GetVDPreference(     iCD_Drive_Letter : Integer;
                                    var Preference       : Integer ) : Boolean;
var regs : Registers;

Begin                                                   { AX = $150D }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_GETSET_VD_PREFERENCE;
  regs.cx := iCD_Drive_Letter;
  regs.bx := $0000;                             { 0 = Get Preference }
  regs.dx := 0;               { Ergebnisregister vorsorglich lschen }
  intr( DOS_MULTIPLEX_INT, regs );
  Preference := regs.dx;
  if ( regs.flags and 1 ) = 0 then MSCDEX200_GetVDPreference := TRUE
                              else MSCDEX200_GetVDPreference := FALSE;
End;

{********************************************************************}
{ MSCDEX200_SetVDPreference : Setze Volume Descriptor Preference     }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           iPreference      - Neue Preference                       }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{********************************************************************}
Function MSCDEX200_SetVDPreference( iCD_Drive_Letter,
                                    iPreference : Integer ) : Boolean;
var regs : Registers;

Begin                                                   { AX = $150D }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_GETSET_VD_PREFERENCE;
  regs.cx := iCD_Drive_Letter;
  regs.bx := $0001;                             { 1 = Set Preference }
  regs.dx := iPreference;
  intr( DOS_MULTIPLEX_INT, regs );
  if ( regs.flags and 1 ) = 0 then MSCDEX200_SetVDPreference := TRUE
                              else MSCDEX200_SetVDPreference := FALSE;
End;

{********************************************************************}
{ MSCDEX200_GetDirectoryEntry : Ermittle Datei-Informationen         }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Laufwerkskennung (0=A, 1=B, 2=C etc.) }
{           iCopyFlag        - 0 - direct copy, 1 - copy to struct   }
{           PathName         - Pfadname der Datei                    }
{           lpData           - Adresse, an die Daten kopiert werden  }
{                              sollen                                }
{           HSG              - Directory Daten HSG(<>0) oder ISO(=0) }
{ Ausgabe : TRUE  - Operation erfolgreich                            }
{           FALSE - Operation fehlgeschlagen ( Fehler in _doserrno ) }
{*------------------------------------------------------------------*}
{ Info : Der bereitgestellte Buffer sollte mindestens 255 Bytes      }
{        (copy to struct) oder 280 Bytes (direct copy) gro sein.    }
{        Bei direct copy ist der Programmierer fr die Dekodierung   }
{        der Strukturdaten verantwortlich. Um welchen der beiden     }
{        Typen (HSG_ENTRY oder ISO_ENTRY) es sich handelt, wird in   }
{        der Variablen vermerkt, die fr den Parameter HSG ber-     }
{        geben wird.                                                 }
{********************************************************************}
Function MSCDEX200_GetDirectoryEntry(     iCD_Drive_Letter : Integer;
                                          iCopyFlag        : Boolean;
                                          PathName         : String;
                                          lpData           : Pointer;
                                      var HSG              : Integer ) : Boolean;
var regs : Registers;

Begin                   { AX = $150F, CL = CD_Drive, CH = copy flags }
                             { ES:BX = Pfadname, SI:DI = Datenadress }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX200_GET_DIRECTORY_ENTRY;
  regs.cl := Byte ( iCD_Drive_Letter );
                                              { Bit 0 setzen/lschen }
  if icopyFlag then regs.ch := $01 else regs.ch := $00;
  Pathname := Pathname+#0;
  regs.es := Longint ( @PathName ) shr 16;
  regs.bx := Longint ( @PathName ) and $ffff + 1;
  regs.si := Longint ( lpData ) shr 16;
  regs.di := Longint ( lpData ) and $ffff;
  intr( DOS_MULTIPLEX_INT, regs );
  HSG := regs.ax;
  if ( regs.flags and 1 ) = 0 then MSCDEX200_GetDirectoryEntry := TRUE
                              else MSCDEX200_GetDirectoryEntry := FALSE;
End;

{********************************************************************}
{ MSCDEX210_SendDeviceRequest: Sende Device-Request an Device-       }
{                              Treiber                               }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Nummer des Laufwerks, an dessen       }
{                              Device-Treiber ein Device-Request ge- }
{                              sendet werden soll. (0=A, 1=B, etc.)  }
{           lpReqHdr         - Adresse des Device-Request-Headers    }
{ Ausgabe : Status des Device nach Operation                         }
{           (s. RequestHeader.wStatus)                               }
{*------------------------------------------------------------------*}
{ Info: Durch diesen Aufruf ruft MSCDEX die Strategy-Routine des     }
{       Device-Treibers auf.                                         }
{********************************************************************}
Function MSCDEX210_SendDeviceRequest(     iCD_Drive_Letter : Integer;
                                      var ReqHdr           : RequestHeader ) : Integer;
var regs : Registers;
Begin                          { AX = $1510, CX = Laufwerksbuchstabe }
                               { ES:BX = Adresse des Request-Headers }

  regs.ax := MSCDEX_MULTIPLEX_CODE or MSCDEX210_SEND_DEVICE_REQUEST;
  regs.cx := iCD_Drive_Letter;
  regs.es := Longint ( @ReqHdr ) shr 16;
  regs.bx := Longint ( @ReqHdr ) and $ffff;
  intr( DOS_MULTIPLEX_INT, regs );
  MSCDEX210_SendDeviceRequest := ReqHdr.wStatus;
End;

{********************************************************************}
{ _CallStrategy: Sende Device-Request an Device-Treiber              }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Nummer des Laufwerks, an dessen       }
{                              Device-Treiber ein Device-Request ge- }
{                              sendet werden soll. (0=A, 1=B, etc.)  }
{           ReqHdr           - der Device-Request-Header             }
{ Ausgabe : Status des Device nach Operation                         }
{           (s. RequestHeader.wStatus)                               }
{*------------------------------------------------------------------*}
{ Info: Diese Funktion ruft zuerst die Strategy, dann die Interrupt- }
{       Funktion des Device-Treibers auf, dessen Header und SubUnit  }
{       durch den bergebenen Laufwerksbuchstaben bestimmt werden.   }
{       Diese Funktion wird aufgerufen, wenn die Versionsnummer von  }
{       MSCDEX < 2.10 ist.                                           }
{********************************************************************}
Function _CallStrategy(     iCD_Drive_Letter : Integer;
                        var ReqHdr           : RequestHeader ) : Integer;

var lpDevHdr : DevHeadPtr;
    dummy    : boolean;

Begin
  if iNumUnits = -1 then { Alle statischen Variablen initialisiert ? }
  Begin
    dummy := MSCDEX_GetNumberOfDriveLetters( iNumUnits, iStart );
    dummy := MSCDEX_GetCDRomDriveDeviceList( @DevElements[0] );

    if iVersion >= $200 then     { Existiert GetCDRomDriveLetters()? }
      dummy := MSCDEX200_GetCDRomDriveLetters( @cCDLetters )
    else         { ansonsten aufeinanderfolgende Laufwerksbuchstaben }
      for i := 0 to iNumUnits - 1 do
        cCDLetters[i] := Byte ( iStart + i );
  End;

  for i := 0 to iNumUnits - 1 do
    if ( cCDLetters[i] = iCD_Drive_Letter ) then  { Welches Laufwerk? }
      Begin
        lpDevHdr := DevElements[i].lpDeviceHeader;   { DeviceHader und }
        ReqHdr.bSubUnit := DevElements[i].bSubUnit;        { SubUnit }
        if ( Longint ( lpDevHdr ) and $ffff ) <> 0 then
          Begin
            _CallStrategy:=-1;                       { Immer an Offset 0 }
            Exit;
         End;
      Strategy := Pointer ( ( Longint ( lpDevHdr ) and $ffff0000 )
                  + lpDevHdr^.wStrategy );
      Interruptproc := Pointer ( ( Longint ( lpDevHdr ) and $ffff0000 )
                       + lpDevHdr^.wInterrupt );
                               { Segment und Offset ermitteln, bevor }
      S := Longint ( @ReqHdr ) shr 16;
                               { asm-Anweisungen ausgefhrt werden!  }
      O := Longint ( @ReqHdr ) and $ffff;
      asm
       mov es,S                                       { ES:BX setzen }
       mov bx,O
       call Strategy       { Zuerst Interne Bufferadressen setzen... }
       call Interruptproc;                  { ...dann Code ausfhren }
      End;
      _CallStrategy := ReqHdr.wStatus;
      Exit;
    End;
  _CallStrategy := -1;
End;

{********************************************************************}
{ MSCDEX_SendDeviceRequest: Sende Device-Request an Device-          }
{                           Treiber                                  }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter - Nummer des Laufwerks, an dessen       }
{                              Device-Treiber ein Device-Request ge- }
{                              sendet werden soll. (0=A, 1=B, etc.)  }
{           ReqHdr           - Device-Request-Header                 }
{ Ausgabe : Status des Device nach Operation                         }
{           (s. RequestHeader.wStatus)                               }
{*------------------------------------------------------------------*}
{ Info: Dise Funktion ruft die Interrupt- und Strategy-Routinen ent- }
{       weder ber den MSCDEX-MUX auf, oder ruft _CallStrategy auf.  }
{       Auf welchem Weg die Device-Treiber funktionen aufgerufen     }
{       werden ist abhngig von der MSCDEX-Version.                  }
{********************************************************************}
Function MSCDEX_SendDeviceRequest(     iCD_Drive_Letter : Integer;
                                   var ReqHdr           : RequestHeader ) : Integer;
Begin
  if iVersion = 0 then  { Falls Version noch nicht bekannt, ermitteln }
    iVersion := MSCDEX200_GetVersion;

  if iVersion < $210 then  { Bei Versionen kleiner 2.10 Direktaufruf }
    MSCDEX_SendDeviceRequest :=
      _CallStrategy( iCD_Drive_Letter, ReqHdr )
  else
    MSCDEX_SendDeviceRequest :=
      MSCDEX210_SendDeviceRequest( iCD_Drive_Letter, ReqHdr );
End;

{********************************************************************}
{ MSCDEX_ReadWriteReq : Fhre MSCDEX-IOCTLI_READ/WRITE-Request aus   }
{*------------------------------------------------------------------*}
{ Eingabe : iCD_Drive_Letter  - Kennung des Laufwerks, an dessen     }
{                               Treiber ein Request abgeschickt      }
{                               werden soll.                         }
{           iCommand          - Nummer des auszufhrenden Kommandos  }
{           lpControlBlock    - Adresse des Controlblocks, der READ/ }
{                               WRITE-Request nher bestimmt.        }
{           iControlBlockSize - Gre des ControlBlocks              }
{ Ausgabe : Device-Status                                            }
{*------------------------------------------------------------------*}
{ Info    : MSCDEX fllt das SubUnit-Feld bei Aufruf eines Treibers, }
{           der mehrere Laufwerke untersttzt, automatisch aus. Wird }
{           ein Treiber-Kommando nicht mit Hilfe von MSCDEX ab-      }
{           gesetzt, mu das SubUnit Feld vom Programmierer gesetzt  }
{           werden (s. _CallStrategy() )                             }
{********************************************************************}
Function MSCDEX_ReadWriteReq( iCD_Drive_Letter,
                              iCommand          : Integer;
                              lpControlBlock    : Pointer;
                              iControlBlockSize : Integer ) : Integer;

var IOCTLIO : MSCDEX_IOCTLIO;

Begin
  IOCTLIO.ReqHdr.bLength    := sizeof( RequestHeader );
  IOCTLIO.ReqHdr.bSubUnit   := 0;
  IOCTLIO.ReqHdr.bCommand   := Byte ( iCommand );
  IOCTLIO.bMediaDescriptor  := 0;
  IOCTLIO.lpTransferAdress  := lpControlBlock;
  IOCTLIO.wTransferCount    := iControlBlockSize;
  IOCTLIO.wStartingSectorNo := 0;
  IOCTLIO.lpVolumeID        := NIL;

  MSCDEX_ReadWriteReq:=MSCDEX_SendDeviceRequest( iCD_Drive_Letter,
                                                 IOCTLIO.ReqHdr );
End;

End.
